/*******************************************************************************
 DataModule.cc

	DataModule class.
 
	Copyright @ 1997 by Glenn W. Bach. All rights reserved.

 ******************************************************************************/
 
#include "DataModule.h"
#include "GXRaggedFloatTable.h"
#include "GRaggedFloatTableData.h"
#include "GloveModule.h"
#include "GLUndoElementsInsert.h"
#include <DeleteBroadcasterTask.h>

#include <JXApplication.h>

#include <JInPipeStream.h>
#include <JOutPipeStream.h>
#include <JProcess.h>
//#include <JIPCLine.h>
#include <JUserNotification.h>
#include <JCreateProgressDisplay.h>
#include <JError.h>
#include <JString.h>

#include <jXGlobals.h>

#include <jProcessUtil.h>
#include <jStreamUtil.h>

#include <strstream.h>
#include <iostream.h>
#include <jAssert.h>

#undef new
#undef delete

/*********************************************************************************
 Create
 

 ********************************************************************************/

JBoolean
DataModule::Create
	(
	DataModule** module,
	GXRaggedFloatTable* table, 
	GRaggedFloatTableData* data,
	const JCharacter* sysCmd
	)
{
	int inFD;
	int outFD;
	JProcess* process;
	JError err = 
			JProcess::Create(&process, sysCmd,
							kJCreatePipe, &outFD,
							kJCreatePipe, &inFD,
							kJIgnoreConnection, NULL);	

	if (err.OK())
		{
		JInPipeStream* ip = new JInPipeStream(inFD, kTrue);
		assert( ip != NULL );
		assert( ip->good() );
		JOutPipeStream* op = new JOutPipeStream(outFD, kTrue);
		assert( op != NULL );
		assert( op->good() );
		*module = new DataModule(table, data, process, ip, op, inFD);
		return kTrue;
		}
		
	return kFalse;
}

/*******************************************************************************
 Constructor (protected)
 

 ******************************************************************************/

DataModule::DataModule
	(
	GXRaggedFloatTable* 	table, 
	GRaggedFloatTableData*	data,
	JProcess* 				process, 
	JInPipeStream* 			input, 
	JOutPipeStream* 		output,
	const int				fd
	)
	:
	JBroadcaster()
{
	itsTable = table;
	itsData = data;
	ListenTo(itsData);
	itsProcess = process;
	ListenTo(itsProcess);

	itsLink = new ProcessLink(fd);
//	itsLink->set_hanle(input);
	assert(itsLink != NULL);
	ListenTo(itsLink);

	itsInput = input;
//	ListenTo(input);
	itsOutput = output;
//	ListenTo(output);
	itsHeaderRead = kFalse;
	itsSentData = kFalse;
	itsPG = NULL;
	itsCols = new JArray<JIndex>;
	assert(itsCols != NULL);
}

/*******************************************************************************
 Destructor
 

 ******************************************************************************/

DataModule::~DataModule()
{
	delete itsProcess;
//	delete itsProcessInput;
	delete itsLink;
	if (itsPG != NULL)
		{
		itsPG->ProcessFinished();
		delete itsPG;
		}
}

/*******************************************************************************
 Receive
 

 ******************************************************************************/

void
DataModule::Receive
	(
	JBroadcaster* sender,
	const JBroadcaster::Message& message
	)
{
	if (message.Is(JProcess::kFinished))
		{
		JGetUserNotification()->ReportError("Data module completed.");
		}
	else if (sender == itsLink && message.Is(JMessageProtocolT::kMessageReady))
		{
		if (itsLink->HasMessages())
			{
			JSize count = itsLink->GetMessageCount();
			for (JSize i = 1; i <= count; i++)
				{
				JString str;
				if (itsLink->GetNextMessage(&str))
					{
					HandleInput(str);
					}
				}
			}
		}
	else if (sender == itsData && message.Is(JTableData::kColRemoved))
		{
		const JTableData::ColRemoved* info =
			dynamic_cast(const JTableData::ColRemoved*, &message);
		assert( info != NULL );
		JIndex index = info->GetIndex();
		for (JIndex i = 1; i <= itsCols->GetElementCount(); i++)
			{
			JIndex cindex = itsCols->GetElement(i);
			if (index < cindex)
				{
				itsCols->SetElement(i, cindex-1);
				}
			else if (index == cindex)
				{
				JGetUserNotification()->ReportError("Module column removed.");
				DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
				assert(dbt != NULL);
				JXGetApplication()->InstallUrgentTask(dbt);
				}
			}
		}
		
	else if (sender == itsData && message.Is(JTableData::kColInserted))
		{
		const JTableData::ColInserted* info =
			dynamic_cast(const JTableData::ColInserted*, &message);
		assert( info != NULL );
		JIndex index = info->GetIndex();
		for (JIndex i = 1; i <= itsCols->GetElementCount(); i++)
			{
			JIndex cindex = itsCols->GetElement(i);
			if (index <= cindex)
				{
				itsCols->SetElement(i, cindex+1);
				}
			}
		}
}

/*******************************************************************************
 HandleInput
 

 ******************************************************************************/

void
DataModule::HandleInput
	(
	const JString& str
	)
{
	if (itsHeaderRead)
		{
		HandleDataRead(str);
		}
	else
		{
		istrstream iss(str);
		int eval;
		if (!itsSentData)
			{
			iss >> eval;
			iss >> ws;
			if (eval == kGloveRequiresData)
				{
				int cols;
				iss >> cols;
				iss >> ws;
				JBoolean success = itsTable->WriteDataCols(*itsOutput, cols);
				if (!success)
					{
					JGetUserNotification()->ReportError("Module Error.");
					DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
					assert(dbt != NULL);
					JXGetApplication()->InstallUrgentTask(dbt);
					return;
					}
				itsSentData = kTrue;
				return;
				}
			}
		iss >> eval;
		iss >> ws;
		if (eval == kGloveDataDump)
			{
			itsDataIsDump = kTrue;
			}
		else
			{
			itsDataIsDump = kFalse;
			}
		iss >> eval;
		iss >> ws;
		HandlePrepareCols(eval);
		itsHeaderRead = kTrue;
		}
}

/*******************************************************************************
 HandlePrepareCols
 

 ******************************************************************************/

void
DataModule::HandlePrepareCols
	(
	const JSize cols
	)
{
	itsColNum = cols;
	itsColStart = itsData->GetDataColCount() + 1;
	for (JSize i = 1; i <= cols; i++)
		{
		itsData->AppendCol();
		itsCols->AppendElement(itsColStart - 1 + i);
		}
	GLUndoElementsInsert* undo =
		new GLUndoElementsInsert(itsTable, JPoint(itsColStart, 1), 
								 JPoint(itsColStart + itsColNum - 1, itsTable->GetRowCount()),
								 GLUndoElementsBase::kCols);
	assert(undo != NULL);
	itsTable->NewUndo(undo);
}

/*******************************************************************************
 HandleDataRead
 

 ******************************************************************************/

void
DataModule::HandleDataRead
	(
	const JString& str
	)
{
	if (itsPG == NULL)
		{
		itsPG = JGetCreatePG()->New();
		itsPG->VariableLengthProcessBeginning(
			"Loading data...", kTrue, kTrue);
		itsCurrentRow = 0;
		}
	istrstream iss(str);
	int eval;
	iss >> eval;
	if (eval == kGloveFinished)
		{
		DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
		assert(dbt != NULL);
		JXGetApplication()->InstallUrgentTask(dbt);
		return;
		}
	else if (eval == kGloveFail)
		{
		JGetUserNotification()->ReportError("Module Error.");
		DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
		assert(dbt != NULL);
		JXGetApplication()->InstallUrgentTask(dbt);
		return;
		}
	else if (eval != kGloveOK)
		{
		JGetUserNotification()->ReportError("Module Error.");
		DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
		assert(dbt != NULL);
		JXGetApplication()->InstallUrgentTask(dbt);
		return;
		}
	iss >> ws;
	JFloat value;
	if (itsDataIsDump)
		{
		for (JSize i = 0; i < itsColNum; i++)
			{
			if (itsInput->good())
				{
				iss >> value;
				iss >> ws;
				JIndex cindex = itsCols->GetElement(i+1);
				assert(itsData->ColumnValid(cindex));
				itsData->AppendElement(cindex, value);
				}
			else
				{
				JGetUserNotification()->ReportError("Module Error.");
				DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
				assert(dbt != NULL);
				JXGetApplication()->InstallUrgentTask(dbt);
				return;		
				}
			}
		}
	else
		{
		JSize row, col;
		if (itsInput->good())
			{
			iss >> row;
			iss >> ws;
			if (itsInput->good())
				{
				iss >> col;
				iss >> ws;
				if (itsInput->good())
					{
					iss >> value;
					if (col <= itsColNum)
						{
						JIndex cindex = itsCols->GetElement(col);
						assert(itsData->CellValid(row,cindex));
						itsData->SetElement(row, cindex, value);
						}
					else
						{
						JGetUserNotification()->ReportError("Module Error.");
						DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
						assert(dbt != NULL);
						JXGetApplication()->InstallUrgentTask(dbt);
						return;	
						}
					}
				else
					{
					JGetUserNotification()->ReportError("Module Error.");
					DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
					assert(dbt != NULL);
					JXGetApplication()->InstallUrgentTask(dbt);
					return;	
					}
				}
			else
				{
				JGetUserNotification()->ReportError("Module Error.");
				DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
				assert(dbt != NULL);
				JXGetApplication()->InstallUrgentTask(dbt);
				return;	
				}
			}
		else
			{
			JGetUserNotification()->ReportError("Module Error.");
			DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
			assert(dbt != NULL);
			JXGetApplication()->InstallUrgentTask(dbt);
			return;		
			}
		}
	itsCurrentRow++;
	JBoolean keepGoing = itsPG->IncrementProgress(itsCurrentRow);
	if (!keepGoing)
		{
		DeleteBroadcasterTask* dbt = new DeleteBroadcasterTask(this);
		assert(dbt != NULL);
		JXGetApplication()->InstallUrgentTask(dbt);
		return;
		}
	itsTable->Redraw();
}
