/******************************************************************************
 GRaggedFloatTableData.cc

	class for storing floats in a ragged table.

	BASE CLASS = JTableData

	Copyright  1997 by Glenn W. Bach. All rights reserved.

 ******************************************************************************/

#include "GRaggedFloatTableData.h"
#include <JMinMax.h>
#include <jAssert.h>

const JCharacter* GRaggedFloatTableData::kElementInserted = "GRaggedFloatTableData::ElementInserted";
const JCharacter* GRaggedFloatTableData::kElementRemoved = "GRaggedFloatTableData::ElementRemoved";
const JCharacter* GRaggedFloatTableData::kDataChanged = "GRaggedFloatTableData::DataChanged";

/******************************************************************************
 Constructor

 ******************************************************************************/

GRaggedFloatTableData::GRaggedFloatTableData
	(
	const JFloat& defValue
	)
	:
	JTableData(),
	itsDefValue(defValue)
{
	itsCols = new JPtrArray< JArray<JFloat> >;
	assert( itsCols != NULL );
	
	const JSize kInitialColCount = 20;
	const JSize kInitialRowCount = 100;
	for (JSize i = 1; i <= kInitialColCount; i++)
		{
		ColAdded();
		Broadcast(JTableData::ColInserted(1));
		}
	for (JSize i = 1; i <= kInitialRowCount; i++)
		{
		RowAdded();
		Broadcast(JTableData::RowInserted(1));
		}
	itsBroadcast = kTrue;
}

/******************************************************************************
 Copy constructor

 ******************************************************************************/

GRaggedFloatTableData::GRaggedFloatTableData
	(
	const GRaggedFloatTableData& source
	)
	:
	JTableData(source),
	itsDefValue(source.itsDefValue),
	itsBroadcast(source.itsBroadcast)
{
	itsCols = new JPtrArray< JArray<JFloat> >;
	assert( itsCols != NULL );

	const JSize count = (source.itsCols)->GetElementCount();
	for (JIndex i=1; i<=count; i++)
		{
		const JArray<JFloat>* origColData = (source.itsCols)->NthElement(i);
		JArray<JFloat>* newColData = new JArray<JFloat>(*origColData);
		assert( newColData != NULL );
		itsCols->Append(newColData);
		}
}

/******************************************************************************
 Destructor

 ******************************************************************************/

GRaggedFloatTableData::~GRaggedFloatTableData()
{
	itsCols->DeleteAll();
	delete itsCols;
}

/******************************************************************************
 CellValid

 ******************************************************************************/

JBoolean
GRaggedFloatTableData::CellValid
	(
	const JIndex row,
	const JIndex col
	)
	const
{
	const JSize colCount = itsCols->GetElementCount();
	if (col > colCount)
		{
		return kFalse;
		}

	const JArray<JFloat>* dataCol = itsCols->NthElement(col);
	const JSize rowCount = dataCol->GetElementCount();
	if (row > rowCount)
		{
		return kFalse;
		}	
	return kTrue;
}

JBoolean
GRaggedFloatTableData::CellValid
	(
	const JPoint& cell
	)
	const
{
	return CellValid(cell.y, cell.x);
}

/******************************************************************************
 GetElement

 ******************************************************************************/

JBoolean
GRaggedFloatTableData::GetElement
	(
	const JIndex row,
	const JIndex col,
	JFloat* value
	)
	const
{
	if (CellValid(row,col) == kFalse)
		{
		return kFalse;
		}
		
	const JArray<JFloat>* dataCol = itsCols->NthElement(col);
	*value =  dataCol->GetElement(row);
	return kTrue;
}

JBoolean
GRaggedFloatTableData::GetElement
	(
	const JPoint& cell, 
	JFloat* value
	)
	const
{
	if (CellValid(cell) == kFalse)
		{
		return kFalse;
		}
		
	const JArray<JFloat>* dataCol = itsCols->NthElement(cell.x);
	*value =  dataCol->GetElement(cell.y);
	return kTrue;
}

/******************************************************************************
 CreateCellIfNeeded

 ******************************************************************************/

void
GRaggedFloatTableData::CreateCellIfNeeded
	(
	const JIndex row,
	const JIndex col
	)
{
	JSize colCount = itsCols->GetElementCount();
	while (col > colCount)
		{
		AppendCol();
		colCount = itsCols->GetElementCount();
		}
		
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	JSize rowCount = dataCol->GetElementCount();
	while (row > rowCount)
		{
		AppendElement(col, itsDefValue);
		rowCount = dataCol->GetElementCount();
		}
}

void
GRaggedFloatTableData::CreateCellIfNeeded
	(
	const JPoint cell
	)
{
	CreateCellIfNeeded(cell.y, cell.x);
}


/******************************************************************************
 SetElement

 ******************************************************************************/

void
GRaggedFloatTableData::SetElement
	(
	const JIndex	row,
	const JIndex	col,
	const JFloat&		data
	)
{
	CreateCellIfNeeded(row, col);	
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	dataCol->SetElement(row, data);
	CheckBroadcast(JTableData::RectChanged(row,col));
}

void
GRaggedFloatTableData::SetElement
	(
	const JPoint&	cell1,
	const JFloat&		data
	)
{
	JPoint cell = cell1;
	JCoordinate row = cell.y;
	JCoordinate col = cell.x;
	CreateCellIfNeeded(row, col);
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	dataCol->SetElement(row, data);
	CheckBroadcast(JTableData::RectChanged(cell));
}

/******************************************************************************
 GetRow

 ******************************************************************************/

void
GRaggedFloatTableData::GetRow
	(
	const JIndex			index,
	JArray<JFloat>*	rowData
	)
	const
{
	rowData->RemoveAll();

	const JSize colCount = GetDataColCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		const JArray<JFloat>* dataCol = itsCols->NthElement(i);
		rowData->AppendElement(dataCol->GetElement(index));
		}
}

/******************************************************************************
 SetRow

 ******************************************************************************/

void
GRaggedFloatTableData::SetRow
	(
	const JIndex				index,
	const JArray<JFloat>&	rowData
	)
{
	const JSize colCount = GetDataColCount();
	assert( rowData.GetElementCount() == colCount );

	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* dataCol = itsCols->NthElement(i);
		dataCol->SetElement(index, rowData.GetElement(i));
		}

	Broadcast(JTableData::RectChanged(JRect(index, 1, index+1, colCount+1)));
}

/******************************************************************************
 GetCol

	operator= doesn't work because it could be -any- class derived from JOrderedSet.

 ******************************************************************************/

void
GRaggedFloatTableData::GetCol
	(
	const JIndex	index,
	JOrderedSet<JFloat>*	colData
	)
	const
{
	colData->RemoveAll();

	const JArray<JFloat>* dataCol = itsCols->NthElement(index);
	const JSize rowCount = dataCol->GetElementCount();
	
	for (JIndex i=1; i<=rowCount; i++)
		{
		colData->InsertElementAtIndex(i, dataCol->GetElement(i));
		}
}

/******************************************************************************
 GetCol

	operator= doesn't work because it could be -any- class derived from JOrderedSet.

 ******************************************************************************/

JArray<JFloat>* 
GRaggedFloatTableData::GetColPointer
	(
	const JIndex index
	) 
	const
{
	assert(itsCols->IndexValid(index));
	return itsCols->NthElement(index);
}

/******************************************************************************
 SetCol

	operator= doesn't work because it could be -any- class derived from JOrderedSet.

 ******************************************************************************/

void
GRaggedFloatTableData::SetCol
	(
	const JIndex			index,
	const JOrderedSet<JFloat>&	colData
	)
{
	JArray<JFloat>* dataCol = itsCols->NthElement(index);
	const JSize rowCount = colData.GetElementCount();
	CreateCellIfNeeded(rowCount, index);
	
	for (JIndex i=1; i<=rowCount; i++)
		{
		dataCol->SetElement(i, colData.GetElement(i));
		}

	JRect rect(1, index, rowCount + 1, index + 1);
	CheckBroadcast(JTableData::RectChanged(rect));
}

/******************************************************************************
 InsertRow

 ******************************************************************************/

void
GRaggedFloatTableData::InsertRow
	(
	const JIndex				index,
	const JOrderedSet<JFloat>*	initData
	)
{
	assert( initData == NULL || initData->GetElementCount() == GetDataColCount() );

	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* colData = itsCols->NthElement(i);
		if (index <= colData->GetElementCount())
			{
			JFloat value = 0;
			if ((initData != NULL) && (i <= initData->GetElementCount()))
				{
				value = initData->GetElement(i);
				}
			InsertElement(index, i);
			}
		}
		
/*	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* colData = itsCols->NthElement(i);
		if (index <= colData->GetElementCount())
			{
			colData->InsertElementAtIndex(index, itsDefValue);
			}
		}

	if (GetMaxRowCount() == GetRowCount())
		{
		RowAdded();
		CheckBroadcast(JTableData::RowInserted(index));
		}*/
}

/******************************************************************************
 Row manipulations

 ******************************************************************************/

void
GRaggedFloatTableData::PrependRow()
{
	InsertRow(1);
}

/******************************************************************************
 DuplicateRow

 ******************************************************************************/

void
GRaggedFloatTableData::DuplicateRow
	(
	const JIndex index
	)
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		
		if (index <= rowCount)
			{
			const JFloat element = colData->GetElement(index);
			colData->InsertElementAtIndex(index, element);
			}
		}

	RowAdded();
	CheckBroadcast(JTableData::RowDuplicated(index, index));
}

/******************************************************************************
 RemoveRow

 ******************************************************************************/

void
GRaggedFloatTableData::RemoveRow
	(
	const JIndex index
	)
{
	assert( index < GetRowCount() );
	
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		
		if (index <= rowCount)
			{
			RemoveElement(index, i);
			}
		}

}

/******************************************************************************
 RemoveAllRows

 ******************************************************************************/

void
GRaggedFloatTableData::RemoveAllRows()
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* colData = itsCols->NthElement(i);
		colData->RemoveAll();
		}

	SetRowCount(0);
	CheckBroadcast(JTableData::AllRowsRemoved());
	RowAdded();
	CheckBroadcast(JTableData::RowInserted(1));
}

/******************************************************************************
 MoveRow

 ******************************************************************************/

void
GRaggedFloatTableData::MoveRow
	(
	const JIndex origIndex,
	const JIndex newIndex
	)
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JArray<JFloat>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		if ( (origIndex <= rowCount) && (newIndex <= rowCount) )
			{
			colData->MoveElementToIndex(origIndex, newIndex);
			}
		}

	CheckBroadcast(JTableData::RowMoved(origIndex, newIndex));
}

/******************************************************************************
 InsertCol

 ******************************************************************************/

void
GRaggedFloatTableData::InsertCol
	(
	const JIndex			index,
	const JOrderedSet<JFloat>*	initData
	)
{
	JIndex trueIndex = index;
	const JIndex maxIndex = itsCols->GetElementCount()+1;
	if (trueIndex > maxIndex)
		{
		trueIndex = maxIndex;
		}

	JArray<JFloat>* colData = new JArray<JFloat>;
	assert( colData != NULL );
	itsCols->InsertAtIndex(trueIndex, colData);

	if (initData != NULL)
		{
		const JSize rowCount = initData->GetElementCount();
		for (JIndex i=1; i<=rowCount; i++)
			{
			colData->InsertElementAtIndex(i, initData->GetElement(i));
			}
		}

	ColAdded();
	CheckBroadcast(JTableData::ColInserted(trueIndex));
}

/******************************************************************************
 Column manipulations

 ******************************************************************************/

void
GRaggedFloatTableData::PrependCol
	(
	const JOrderedSet<JFloat>* initData
	)
{
	InsertCol(1, initData);
}

void
GRaggedFloatTableData::AppendCol
	(
	const JOrderedSet<JFloat>* initData
	)
{
	InsertCol(itsCols->GetElementCount()+1, initData);
}

/******************************************************************************
 DuplicateCol

 ******************************************************************************/

void
GRaggedFloatTableData::DuplicateCol
	(
	const JIndex origIndex,
	const JIndex newIndex
	)
{
	JIndex trueIndex = newIndex;
	const JIndex maxIndex = itsCols->GetElementCount()+1;
	if (trueIndex > maxIndex)
		{
		trueIndex = maxIndex;
		}

	JArray<JFloat>* origColData = itsCols->NthElement(origIndex);
	JArray<JFloat>* newColData = new JArray<JFloat>(*origColData);
	assert( newColData != NULL );
	itsCols->InsertAtIndex(trueIndex, newColData);

	ColAdded();
	CheckBroadcast(JTableData::ColDuplicated(origIndex, trueIndex));
}

/******************************************************************************
 RemoveCol

 ******************************************************************************/

void
GRaggedFloatTableData::RemoveCol
	(
	const JIndex index
	)
{
	assert( ColumnValid(index) );
	if (itsCols->GetElementCount() != 1)
		{
		itsCols->DeleteElement(index);
		}
	else
		{
		RemoveAllElements(index);
		itsCols->DeleteElement(index);
		}

	ColDeleted();
	CheckBroadcast(JTableData::ColRemoved(index));
}

/******************************************************************************
 RemoveAllCols

 ******************************************************************************/

void
GRaggedFloatTableData::RemoveAllCols()
{
	itsCols->DeleteAll();

	SetColCount(0);
	CheckBroadcast(JTableData::AllColsRemoved());
	ColAdded();
	CheckBroadcast(JTableData::ColInserted(1));
}

/******************************************************************************
 MoveCol

 ******************************************************************************/

void
GRaggedFloatTableData::MoveCol
	(
	const JIndex origIndex,
	const JIndex newIndex
	)
{
	itsCols->MoveElementToIndex(origIndex, newIndex);
	CheckBroadcast(JTableData::ColMoved(origIndex, newIndex));
}

/******************************************************************************
 InsertElement

 ******************************************************************************/

void
GRaggedFloatTableData::InsertElement
	(
	const JIndex row, 
	const JIndex col, 
	const JFloat value
	)
{
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	dataCol->InsertElementAtIndex(row, value);
	const JSize rowCount = dataCol->GetElementCount();
	CheckBroadcast(GRaggedFloatTableData::ElementInserted(row, col));
	
	if (rowCount == GetRowCount())
		{
		RowAdded();
		CheckBroadcast(JTableData::RowInserted(rowCount+1));
		}
}

void
GRaggedFloatTableData::InsertElement
	(
	const JPoint& cell, 
	const JFloat value
	)
{
	InsertElement(cell.y, cell.x, value);
}

/******************************************************************************
 PrependElement

 ******************************************************************************/

void
GRaggedFloatTableData::PrependElement
	(
	const JIndex col, 
	const JFloat value
	)
{
	InsertElement(1, col, value);
}

/******************************************************************************
 AppendElement

 ******************************************************************************/

void
GRaggedFloatTableData::AppendElement
	(
	const JIndex col, 
	const JFloat value
	)
{
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	InsertElement(dataCol->GetElementCount()+1, col, value);
}

/******************************************************************************
 DuplicateElement

 ******************************************************************************/

void
GRaggedFloatTableData::DuplicateElement
	(
	const JIndex row, 
	const JIndex col
	)
{
	JFloat value;
	if (GetElement(row,col, &value))
		{
		InsertElement(row, col, value);
		}
}

void
GRaggedFloatTableData::DuplicateElement
	(
	const JPoint& cell
	)
{
	JFloat value;
	if (GetElement(cell, &value))
		{
		InsertElement(cell, value);
		}

}

/******************************************************************************
 RemoveElement

 ******************************************************************************/

void
GRaggedFloatTableData::RemoveElement
	(
	const JIndex row, 
	const JIndex col
	)
{
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	const JSize rowCount = dataCol->GetElementCount();
	if (row <= rowCount)
		{
		dataCol->RemoveElement(row);
		}
	CheckBroadcast(GRaggedFloatTableData::ElementRemoved(row, col));
	
	if (GetMaxRowCount() == GetRowCount() - 2)
		{
		CheckBroadcast(JTableData::RowRemoved(GetRowCount()));
		RowDeleted();
		}
}

void
GRaggedFloatTableData::RemoveElement
	(
	const JPoint& cell
	)
{
	RemoveElement(cell.y, cell.x);
}

/******************************************************************************
 RemoveAllElements

 ******************************************************************************/

void
GRaggedFloatTableData::RemoveAllElements
	(
	const JIndex col
	)
{
	JArray<JFloat>* dataCol = itsCols->NthElement(col);
	const JSize rowCount = dataCol->GetElementCount();
	for (JSize i = 1; i <= rowCount; i++)
		{
		RemoveElement(1, col);
		}
}

/******************************************************************************
 MoveElement

 ******************************************************************************/

void
GRaggedFloatTableData::MoveElement
	(
	const JIndex origRow, 
	const JIndex origCol, 
	const JIndex newRow, 
	const JIndex newCol
	)
{
	JFloat value;
	if (GetElement(origRow, origCol, &value))
		{
		InsertElement(newRow, newCol, value);
		RemoveElement(origRow, origCol);
		}
}

void
GRaggedFloatTableData::MoveElement
	(
	const JPoint& origCell, 
	const JPoint& newCell
	)
{
	JFloat value;
	if (GetElement(origCell, &value))
		{
		InsertElement(newCell, value);
		RemoveElement(origCell);
		}
}

/******************************************************************************
 GetMaxRowCount

 ******************************************************************************/

JSize 
GRaggedFloatTableData::GetMaxRowCount()
{
	JSize max = 0;
	const JSize colCount = itsCols->GetElementCount();
	for (JSize i = 1; i <= colCount; i++)
		{
		JArray<JFloat>* dataCol = itsCols->NthElement(i);
		if (dataCol->GetElementCount() > max)
			{
			max = dataCol->GetElementCount();
			}
		}
	return max;
}

/******************************************************************************
 GetDataColCount

 ******************************************************************************/

JSize 
GRaggedFloatTableData::GetDataColCount()
	const
{
	return itsCols->GetElementCount();
}

/******************************************************************************
 ColumnValid

 ******************************************************************************/

JBoolean 
GRaggedFloatTableData::ColumnValid
	(
	const JIndex index
	)
	const
{
	if (index <= (JIndex)itsCols->GetElementCount())
		{
		return kTrue;
		}
	return kFalse;
}

/******************************************************************************
 GetDataColCount

 ******************************************************************************/

JSize 
GRaggedFloatTableData::GetDataRowCount
	(
	const JIndex index
	)
	const
{
	assert( ColumnValid(index) );
	JArray<JFloat>* dataCol = itsCols->NthElement(index);
	return dataCol->GetElementCount();
}

/******************************************************************************
 FindColumn

 ******************************************************************************/

JBoolean 
GRaggedFloatTableData::FindColumn
	(
	const JArray<JFloat>* array, 
	JIndex *index
	)
	const
{
	return itsCols->Find(array, index);
}

/******************************************************************************
 TurnOnBroadcast

 ******************************************************************************/

void
GRaggedFloatTableData::TurnOnBroadcast
	(
	const JBoolean on
	)
{
	itsBroadcast = on;
}

/******************************************************************************
 CheckBroadcast

 ******************************************************************************/

void
GRaggedFloatTableData::CheckBroadcast
	(
	const Message& message
	)
{
	if (itsBroadcast)
		{
		Broadcast(message);
		}
}