/******************************************************************************
 GRaggedStringTableData.v

	class for storing JStrings in a ragged table.


	BASE CLASS = JTableData

	Copyright  1997 Glenn W. Bach. All rights reserved.

 ******************************************************************************/

#include "GRaggedStringTableData.h"
#include "GRaggedFloatTableData.h"
#include "iostream.h"
#include <jAssert.h>

/******************************************************************************
 Constructor

 ******************************************************************************/


GRaggedStringTableData::GRaggedStringTableData
	(
	const GRaggedFloatTableData*	floatData,
	const int				precision
	)
	:
	JTableData()
{
	itsCols = new JPtrArray< JPtrArray<JString> >;
	assert( itsCols != NULL );
	itsFloatData = floatData;
	ListenTo(itsFloatData);
	itsPrecision = precision;
}

/******************************************************************************
 Copy constructor

 ******************************************************************************/


GRaggedStringTableData::GRaggedStringTableData
	(
	const GRaggedStringTableData& source
	)
	:
	JTableData(source)
{
	itsCols = new JPtrArray< JPtrArray<JString> >;
	assert( itsCols != NULL );

	const JSize colCount = (source.itsCols)->GetElementCount();
	for (JIndex j=1; j<=colCount; j++)
		{
		const JPtrArray<JString>* origColData = (source.itsCols)->NthElement(j);

		JPtrArray<JString>* newColData = new JPtrArray<JString>;
		assert( newColData != NULL );
		itsCols->InsertAtIndex(j, newColData);

		const JSize rowCount = origColData->GetElementCount();
		for (JIndex i=1; i<=rowCount; i++)
			{
			const JString* origElement = origColData->NthElement(i);
			JString* newElement  = new JString(*origElement);
			newColData->InsertAtIndex(i, newElement);
			}
		}
}

/******************************************************************************
 Destructor

 ******************************************************************************/


GRaggedStringTableData::~GRaggedStringTableData()
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		(itsCols->NthElement(i))->DeleteAll();
		}
	itsCols->DeleteAll();
	delete itsCols;
}

/******************************************************************************
 Receive (virtual protected)

 ******************************************************************************/

void
GRaggedStringTableData::Receive
	(
	JBroadcaster*	sender,
	const Message&	message
	)
{
	// element data changed

//	if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
//		 && message.Is(JTableData::kElementChanged))
//		{
//		const JTableData::ElementChanged* info =
//			dynamic_cast(const JTableData::ElementChanged*, &message);
//		assert( info != NULL );
//		UpdateCell(info->GetCell());
//		}

	if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kRectChanged))
		{
		const JTableData::RectChanged* info =
			dynamic_cast(const JTableData::RectChanged*, &message);
		assert( info != NULL );
		JRect rect = info->GetRect();
		if (rect.area() == 1)
			{
			UpdateCell(JPoint(rect.left, rect.top));
			}
		else if (rect.height() == 1)
			{
			UpdateRow(rect.top);
			}
		else
			{
			for (JCoordinate col = rect.left; col < rect.right; col++)
				{
				UpdateCol(col);
				}
			}
		}

	// rows changed

//	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
//		 && message.Is(JTableData::kRowChanged))
//		{
//		const JTableData::RowChanged* info =
//			dynamic_cast(const JTableData::RowChanged*, &message);
//		assert( info != NULL );
//		UpdateRow(info->GetIndex());
//		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kRowInserted))
		{
		const JTableData::RowInserted* info =
			dynamic_cast(const JTableData::RowInserted*, &message);
		assert( info != NULL );
		const JIndex index = info->GetIndex();
		InsertRow(index);
		UpdateRow(index);
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kRowDuplicated))
		{
		const JTableData::RowDuplicated* info =
			dynamic_cast(const JTableData::RowDuplicated*, &message);
		assert( info != NULL );
		DuplicateRow(info->GetOrigIndex());
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kRowRemoved))
		{
		const JTableData::RowRemoved* info =
			dynamic_cast(const JTableData::RowRemoved*, &message);
		assert( info != NULL );
		RemoveRow(info->GetIndex());
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kAllRowsRemoved))
		{
		RemoveAllRows();
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kRowMoved))
		{
		const JTableData::RowMoved* info =
			dynamic_cast(const JTableData::RowMoved*, &message);
		assert( info != NULL );
		MoveRow(info->GetOrigIndex(), info->GetNewIndex());
		}

	// columns changed

//	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
//		 && message.Is(JTableData::kColChanged))
//		{
//		const JTableData::ColChanged* info =
//			dynamic_cast(const JTableData::ColChanged*, &message);
//		assert( info != NULL );
//		UpdateCol(info->GetIndex());
//		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kColInserted))
		{
		const JTableData::ColInserted* info =
			dynamic_cast(const JTableData::ColInserted*, &message);
		assert( info != NULL );
		const JIndex index = info->GetIndex();
		InsertCol(index);
		UpdateCol(index);
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kColDuplicated))
		{
		const JTableData::ColDuplicated* info =
			dynamic_cast(const JTableData::ColDuplicated*, &message);
		assert( info != NULL );
		DuplicateCol(info->GetOrigIndex(), info->GetNewIndex());
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kColRemoved))
		{
		const JTableData::ColRemoved* info =
			dynamic_cast(const JTableData::ColRemoved*, &message);
		assert( info != NULL );
		RemoveCol(info->GetIndex());
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 && message.Is(JTableData::kAllColsRemoved))
		{
		RemoveAllCols();
		}

	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 &&	 message.Is(JTableData::kColMoved))
		{
		const JTableData::ColMoved* info =
			dynamic_cast(const JTableData::ColMoved*, &message);
		assert( info != NULL );
		MoveCol(info->GetOrigIndex(), info->GetNewIndex());
		}
		
	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 &&	 message.Is(GRaggedFloatTableData::kElementInserted))
		{
		const GRaggedFloatTableData::ElementInserted* info =
			dynamic_cast(const GRaggedFloatTableData::ElementInserted*, &message);
		assert( info != NULL );
		InsertElement(info->GetRow(), info->GetCol());
		}
		
	else if (sender == const_cast<GRaggedFloatTableData*>(itsFloatData)
		 &&	 message.Is(GRaggedFloatTableData::kElementRemoved))
		{
		const GRaggedFloatTableData::ElementRemoved* info =
			dynamic_cast(const GRaggedFloatTableData::ElementRemoved*, &message);
		assert( info != NULL );
		RemoveElement(info->GetRow(), info->GetCol());
		}


	// something else

	else
		{
		JTableData::Receive(sender, message);
		}
}

/******************************************************************************
 CellValid

 ******************************************************************************/

JBoolean
GRaggedStringTableData::CellValid
	(
	const JIndex row,
	const JIndex col
	)
	const
{
	const JSize colCount = itsCols->GetElementCount();
	if (col > colCount)
		{
		return kFalse;
		}

	JPtrArray<JString>* dataCol = itsCols->NthElement(col);
	const JSize rowCount = dataCol->GetElementCount();
	if (row > rowCount)
		{
		return kFalse;
		}	
	return kTrue;
}

JBoolean
GRaggedStringTableData::CellValid
	(
	const JPoint& cell
	)
	const
{
	return CellValid(cell.y, cell.x);
}


/******************************************************************************
 GetElement

 ******************************************************************************/


JBoolean
GRaggedStringTableData::GetElement
	(
	const JIndex row,
	const JIndex col,
	JString* str
	)
	const
{
	if (CellValid(row,col) == kFalse)
		{
		return kFalse;
		}
		
	JPtrArray<JString>* colData = itsCols->NthElement(col);
	*str =  *(colData->NthElement(row));
	return kTrue;
}


JBoolean
GRaggedStringTableData::GetElement
	(
	const JPoint& cell,
	JString* str
	)
	const
{
	if (CellValid(cell) == kFalse)
		{
		return kFalse;
		}

	JPtrArray<JString>* colData = itsCols->NthElement(cell.x);
	*str =  *(colData->NthElement(cell.y));
	return kTrue;
}

/******************************************************************************
 SetElement

 ******************************************************************************/


void
GRaggedStringTableData::SetElement
	(
	const JIndex	row,
	const JIndex	col,
	const JString&	data
	)
{
	JPtrArray<JString>* colData = itsCols->NthElement(col);
	JString* element = colData->NthElement(row);
	*element = data;
	Broadcast(JTableData::RectChanged(row,col));
}


void
GRaggedStringTableData::SetElement
	(
	const JPoint&	cell,
	const JString&	data
	)
{
	JPtrArray<JString>* colData = itsCols->NthElement(cell.x);
	JString* element = colData->NthElement(cell.y);
	*element = data;
	Broadcast(JTableData::RectChanged(cell));
}

/******************************************************************************
 GetCol

 ******************************************************************************/


void
GRaggedStringTableData::GetCol
	(
	const JIndex	index,
	JPtrArray<JString>*	colData
	)
	const
{
	colData->DeleteAll();

	const JPtrArray<JString>* dataCol = itsCols->NthElement(index);
	const JSize rowCount = dataCol->GetElementCount();
	for (JIndex i=1; i<=rowCount; i++)
		{
		const JString* origElement = dataCol->NthElement(i);
		JString* newElement = new JString(*origElement);
		assert( newElement != NULL );
		colData->InsertAtIndex(i, newElement);
		}
}

/******************************************************************************
 SetCol

 ******************************************************************************/


void
GRaggedStringTableData::SetCol
	(
	const JIndex		index,
	const JPtrArray<JString>&	colData
	)
{
	JPtrArray<JString>* dataCol = itsCols->NthElement(index);
	const JSize rowCount = dataCol->GetElementCount();
	assert( colData.GetElementCount() == rowCount );

	for (JIndex i=1; i<=rowCount; i++)
		{
		JString* element = dataCol->NthElement(i);
		*element   = *(colData.NthElement(i));
		}
	JRect rect(1, index, rowCount+1, index + 1);
	Broadcast(JTableData::RectChanged(rect));
//	Broadcast(JTableData::ColChanged(index));
}

/******************************************************************************
 InsertRow

 ******************************************************************************/


void
GRaggedStringTableData::InsertRow
	(
	const JIndex		index
	)
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		
		if (index <= rowCount)
			{
			JString* element = new JString();
			assert( element != NULL );
			colData->InsertAtIndex(index, element);
			}
		}

	RowAdded();
	Broadcast(JTableData::RowInserted(index));
}

/******************************************************************************
 Row manipulations

 ******************************************************************************/


void
GRaggedStringTableData::PrependRow()
{
	InsertRow(1);
}

/******************************************************************************
 DuplicateRow

 ******************************************************************************/


void
GRaggedStringTableData::DuplicateRow
	(
	const JIndex index
	)
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount(); 

		if (index <= rowCount)
			{
			JString* origElement = colData->NthElement(index);
			JString* newElement  = new JString(*origElement);
			assert( newElement != NULL );
			colData->InsertAtIndex(index, newElement);
			}
		}

	RowAdded();
	Broadcast(JTableData::RowDuplicated(index, index));
}

/******************************************************************************
 RemoveRow

 ******************************************************************************/


void
GRaggedStringTableData::RemoveRow
	(
	const JIndex index
	)
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		
		if (index <= rowCount)
			{
			colData->DeleteElement(index);
			}
		}

	RowDeleted();
	Broadcast(JTableData::RowRemoved(index));
}

/******************************************************************************
 RemoveAllRows

 ******************************************************************************/


void
GRaggedStringTableData::RemoveAllRows()
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		colData->DeleteAll();
		}

	SetRowCount(0);
	Broadcast(JTableData::AllRowsRemoved());
}

/******************************************************************************
 MoveRow

 ******************************************************************************/


void
GRaggedStringTableData::MoveRow
	(
	const JIndex origIndex,
	const JIndex newIndex
	)
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		if ( (origIndex <= rowCount) && (newIndex <= rowCount) )
			{
			colData->MoveElementToIndex(origIndex, newIndex);
			}
		}

	Broadcast(JTableData::RowMoved(origIndex, newIndex));
}

/******************************************************************************
 InsertCol

 ******************************************************************************/


void
GRaggedStringTableData::InsertCol
	(
	const JIndex		index,
	const JPtrArray<JString>*	initData
	)
{
	JIndex trueIndex = index;
	const JIndex maxIndex = itsCols->GetElementCount()+1;
	if (trueIndex > maxIndex)
		{
		trueIndex = maxIndex;
		}

	JPtrArray<JString>* colData = new JPtrArray<JString>;
	assert( colData != NULL );
	itsCols->InsertAtIndex(trueIndex, colData);

	if (initData != NULL)
		{
		const JSize rowCount = initData->GetElementCount();
		for (JIndex i=1; i<=rowCount; i++)
			{
			JString* element = new JString(*(initData->NthElement(i)));
			assert( element != NULL );

			colData->InsertAtIndex(i, element);
			}
		}

	ColAdded();
	Broadcast(JTableData::ColInserted(trueIndex));
}

/******************************************************************************
 Column manipulations

 ******************************************************************************/


void
GRaggedStringTableData::PrependCol
	(
	const JPtrArray<JString>* initData
	)
{
	InsertCol(1, initData);
}


void
GRaggedStringTableData::AppendCol
	(
	const JPtrArray<JString>* initData
	)
{
	InsertCol(itsCols->GetElementCount()+1, initData);
}

/******************************************************************************
 DuplicateCol

 ******************************************************************************/


void
GRaggedStringTableData::DuplicateCol
	(
	const JIndex origIndex,
	const JIndex newIndex
	)
{
	JIndex trueIndex = newIndex;
	const JIndex maxIndex = itsCols->GetElementCount()+1;
	if (trueIndex > maxIndex)
		{
		trueIndex = maxIndex;
		}

	JPtrArray<JString>* origColData = itsCols->NthElement(origIndex);

	JPtrArray<JString>* newColData = new JPtrArray<JString>;
	assert( newColData != NULL );
	itsCols->InsertAtIndex(trueIndex, newColData);

	const JSize rowCount = origColData->GetElementCount();
	for (JIndex i=1; i<=rowCount; i++)
		{
		JString* origElement = origColData->NthElement(i);
		JString* newElement  = new JString(*origElement);
		assert( newElement != NULL );
		newColData->InsertAtIndex(i, newElement);
		}

	ColAdded();
	Broadcast(JTableData::ColDuplicated(origIndex, trueIndex));
}

/******************************************************************************
 RemoveCol

 ******************************************************************************/


void
GRaggedStringTableData::RemoveCol
	(
	const JIndex index
	)
{
	JPtrArray<JString>* colData = itsCols->NthElement(index);
	colData->DeleteAll();

	itsCols->DeleteElement(index);

	ColDeleted();
	Broadcast(JTableData::ColRemoved(index));
}

/******************************************************************************
 RemoveAllCols

 ******************************************************************************/


void
GRaggedStringTableData::RemoveAllCols()
{
	const JSize colCount = itsCols->GetElementCount();
	for (JIndex i=1; i<=colCount; i++)
		{
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		colData->DeleteAll();
		}

	itsCols->DeleteAll();

	SetColCount(0);
	Broadcast(JTableData::AllColsRemoved());
}

/******************************************************************************
 MoveCol

 ******************************************************************************/


void
GRaggedStringTableData::MoveCol
	(
	const JIndex origIndex,
	const JIndex newIndex
	)
{
	itsCols->MoveElementToIndex(origIndex, newIndex);
	Broadcast(JTableData::ColMoved(origIndex, newIndex));
}

/******************************************************************************
 UpdateRow (private)

 ******************************************************************************/

void
GRaggedStringTableData::UpdateRow
	(
	const JIndex index
	)
{
	const JSize count = itsCols->GetElementCount();
	JPoint cell(1,index);
	for (JIndex i=1; i<=count; i++)
		{
		cell.x = i;
		JPtrArray<JString>* colData = itsCols->NthElement(i);
		const JSize rowCount = colData->GetElementCount();
		if (index >= rowCount)
			{
			UpdateCell(cell);
			}
		}
}

/******************************************************************************
 UpdateCol (private)

 ******************************************************************************/

void
GRaggedStringTableData::UpdateCol
	(
	const JIndex index
	)
{
	JPtrArray<JString>* colData = itsCols->NthElement(index);
	const JSize count = colData->GetElementCount();
	JPoint cell(index,1);
	for (JIndex i=1; i<=count; i++)
		{
		cell.y = i;
		UpdateCell(cell);
		}
}

/******************************************************************************
 UpdateCell (private)

 ******************************************************************************/

void
GRaggedStringTableData::UpdateCell
	(
	const JPoint& cell
	)
{
	JFloat value;
	if (itsFloatData->GetElement(cell, &value))
		{
		const JString valueStr(value, itsPrecision);
		SetElement(cell, valueStr);
		}
}

/******************************************************************************
 InsertElement (private)

 ******************************************************************************/

void
GRaggedStringTableData::InsertElement
	(
	const JIndex row, 
	const JIndex col
	)
{
	JPtrArray<JString>* colData = itsCols->NthElement(col);
	JString* str = new JString();
	colData->InsertElementAtIndex(row, str);
	UpdateCell(JPoint(col, row));	
}

/******************************************************************************
 RemoveElement (private)

 ******************************************************************************/

void
GRaggedStringTableData::RemoveElement
	(
	const JIndex row, 
	const JIndex col
	)
{
	JPtrArray<JString>* colData = itsCols->NthElement(col);
	colData->DeleteElement(row);
	UpdateCol(col);	
}

