/*********************************************************************************
 JPlotFitFunction.cc
 
	JPlotFitFunction class.
 
	Copyright @ 1997 by Glenn W. Bach. All rights reserved.

 ********************************************************************************/

#include "JPlotFitFunction.h"
#include "J2DPlotWidget.h"
#include "JPlotDataBase.h"
#include "J2DPlotData.h"
#include "J2DDataPoint.h"
#include "jAssert.h"

/*********************************************************************************
 Constructor 
 

 ********************************************************************************/

JPlotFitFunction::JPlotFitFunction
	(
	J2DPlotWidget* 	plot, 
	JPlotDataBase* 	fitData,
	const JFloat	xMin,
	const JFloat	xMax
	)
	:
	JPlotFunctionBase(JPlotDataBase::kScatterPlot, plot, xMin, xMax),
	JFitBase()
{
	itsData = fitData;
	itsXDiffData = new JArray<JFloat>;
	itsYDiffData = new JArray<JFloat>;
	itsHasXErrors = kFalse;
	itsHasYErrors = kFalse;
	if (itsData->HasXErrors())
		{
		itsXErrDiffData = new JArray<JFloat>;
		itsHasXErrors = kTrue;
		}
	if (itsData->HasYErrors())
		{
		itsHasYErrors = kTrue;
		}
	itsYErrDiffData = new JArray<JFloat>;
}

/*********************************************************************************
 Destructor
 

 ********************************************************************************/
 
JPlotFitFunction::~JPlotFitFunction()
{
	delete itsXDiffData;
	delete itsYDiffData;
	if (itsHasXErrors)
		{
		itsXErrDiffData = new JArray<JFloat>;
		}
	if (itsHasYErrors)
		{
		itsYErrDiffData = new JArray<JFloat>;
		}
}

/*********************************************************************************
 GenerateDiffData (protected)
 

 ********************************************************************************/
 
void
JPlotFitFunction::GenerateDiffData()
{
	itsXDiffData->RemoveAll();
	itsYDiffData->RemoveAll();
	if (!itsData->HasYErrors())
		{
		CalculateStdDev();
		}
	itsYErrDiffData->RemoveAll();
	if (itsData->HasXErrors())
		{
		itsXErrDiffData->RemoveAll();
		}
	const JSize count = itsData->GetElementCount();
	J2DDataPoint data;
	for (JSize i = 1; i <= count; i++)
		{
		if (GetDataElement(i, &data))
			{
			JFloat fitY;
			GetYValue(data.x, &fitY);
			itsXDiffData->AppendElement(data.x);
			itsYDiffData->AppendElement(data.y - fitY);
			if (itsHasYErrors)
				{
				itsYErrDiffData->AppendElement(data.yerr);
				}
			else
				{
				itsYErrDiffData->AppendElement(itsStdDev);
				}
			if (itsHasXErrors)
				{
				itsXErrDiffData->AppendElement(data.xerr);
				}
			}
		}
	J2DPlotData::Create(&itsDiffData, *itsXDiffData, *itsYDiffData, kTrue);
	itsDiffData->SetYErrors(*itsYErrDiffData);
	if (itsHasXErrors)
		{
		itsDiffData->SetXErrors(*itsXErrDiffData);
		}
}

/*********************************************************************************
 GetDiffData
 

 ********************************************************************************/
 
JPlotDataBase*
JPlotFitFunction::GetDiffData()
	const
{
	return itsDiffData;
}

/*********************************************************************************
 CalculateStdDev (private)
 

 ********************************************************************************/
 
void
JPlotFitFunction::CalculateStdDev()
{
	JSize count = itsData->GetElementCount();
	J2DDataPoint data;
	JFloat current = 0;
	for (JSize i = 1; i <= count; i++)
		{
		itsData->GetElement(i, &data);
		JFloat fitY;
		GetYValue(data.x, &fitY);
		current += (fitY - data.y)*(fitY - data.y);
		}
	itsStdDev = sqrt(current/(count - 2));
}

/*********************************************************************************
 AdjustDiffDataValue (protected)
 

 ********************************************************************************/
 
void
JPlotFitFunction::AdjustDiffDataValue
	(
	const JIndex index, 
	const JFloat value
	)
{
	itsYErrDiffData->SetElement(index, value);
}

/*********************************************************************************
 DataElementValid
 

 ********************************************************************************/

JBoolean
JPlotFitFunction::DataElementValid
	(
	const JIndex index
	)
{
	return kTrue;
}

/******************************************************************************
 GetDataElement
 

 *****************************************************************************/

JBoolean
JPlotFitFunction::GetDataElement
	(
	const JIndex index,
	J2DDataPoint* point
	)
{
	JBoolean valid = DataElementValid(index);
	if (!valid)
		{
		return kFalse;
		}
	const JPlotDataBase* data = GetData();
	data->GetElement(index, point);
	return kTrue;
}