/******************************************************************************
 PlotDir.cc

	BASE CLASS = JXWindowDirector

	Copyright  1997 by Glenn Bach. All rights reserved.

 ******************************************************************************/

#include "PlotDir.h"
#include "GVarList.h"
#include "GXPlotFunctionDialog.h"
#include "JPlotLinearFit.h"
#include "JPlotQuadFit.h"
#include "JPlotFitFunction.h"
#include "JFitBase.h"
#include "FitParmsDir.h"
#include "JPlotDataBase.h"
#include "J2DPlotData.h"
#include "J2DVectorData.h"
#include "GRaggedFloatTableData.h"
#include "GloveHistoryDir.h"
#include "FitModule.h"
#include "FitModuleDialog.h"
#include "GXPlotApp.h"
#include "JPlotModuleFit.h"
#include "JPlotJFunction.h"
#include <GlovePlotter.h>
#include "gloveplotdoc.h"
#include <GLGlobals.h>
#include "GLHelpText.h"
#include "GXPlotApp.h"
#include <GLClosePlotDirTask.h>

#include <JXTextMenu.h>
#include <JXWindow.h>
#include <JXMenuBar.h>
#include <JXDocumentMenu.h>
#include <JXFileDocument.h>
#include <JXApplication.h>
#include <JXDialogDirector.h>
#include <JXPSPrinter.h>
#include <JXHelpDirector.h>

#include <JFunction.h>
#include <JFunctionWithVar.h>
#include <JExprNodeList.h>
#include <JFunctionType.h>
#include <JArray.h>
#include <JPtrArray.h>
#include <JString.h>
#include <JUserNotification.h>

#include <jXCSFIcons.h>

#include <jParseFunction.h>

#include <iostream.h>
#include <fstream.h>
#include <jAssert.h>

// setup information

const JFileVersion kCurrentSetupVersion = 1;

extern GXPlotApp* 			gPlotApp;

const JSize kFunctionSteps = 100;

static const JCharacter* kWindowListMenuTitleStr = "Windows";

static const JCharacter* kPlotMenuTitleStr  = "Analysis";
static const JCharacter* kPlotMenuStr = 
	"Page setup... | Print... %k Meta-P | Print session... %l|"
	"Plot function... %l| Linear fit | Quadratic fit | Exponential fit |  Select fit module"
	"%l| Fit parameters | Residual plots"
	"%l| Close %k Meta-W";

enum
{
	kPageSetupCmd = 1,
	kPrintCmd,
	kPrintSessionCmd,
	kPlotFunctionCmd,
	kLinearFitCmd,
	kQuadFitCmd,
	kExpFitCmd,
	kFitModuleCmd,
	kFitParmsCmd,
	kDiffPlotCmd,
	kCloseCmd
};

static const JCharacter* kFitParmsMenuTitleStr = "Fit Parameters";
static const JCharacter* kCurveMenuTitleStr = "Curves";

static const JCharacter* kHelpMenuTitleStr = "Help";
static const JCharacter* kHelpMenuStr = 
	"About %l| Table of Contents | This window "
	"%l|Changes|Credits";

enum
{
	kAboutCmd = 1,
	kTOCCmd,
	kThisWindowCmd,
	kChangesCmd,
	kCreditsCmd
};

/******************************************************************************
 Constructor

 ******************************************************************************/

PlotDir::PlotDir
	(
	JXDirector*       supervisor,
	JXFileDocument*   notifySupervisor,
	const JCharacter* filename,
	const JBoolean    hideOnClose
	)
	:
	JXDocument(supervisor)
{
	itsSupervisor       = notifySupervisor;
	itsFileName         = new JString(filename);
	itsHideOnClose      = hideOnClose;
	itsPlotIsClosing	= kFalse;
	
	itsPrinter          = NULL;

	JCoordinate w 		= 500;
	JCoordinate h       = 400;
	JCoordinate minW	= 400;
	JCoordinate minH    = 300;
	
	JXWindow* window = new JXWindow(this, w,h, filename);
    assert( window != NULL );
    SetWindow(window);
    
    window->SetMinSize(minW,minH);
    if (hideOnClose)
    	{
    	window->SetCloseAction(JXWindow::kDeactivateDirector);
    	}
	
	JXMenuBar* menuBar =
		new JXMenuBar(window,
			JXWidget::kHElastic, JXWidget::kFixedTop, 0,0, w, kJXStdMenuBarHeight);
    assert( menuBar != NULL );
	
	itsPlotMenu = menuBar->AppendTextMenu(kPlotMenuTitleStr);
	itsPlotMenu->SetMenuItems(kPlotMenuStr);
	itsPlotMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsPlotMenu);
	
	itsFitParmsMenu = new JXTextMenu(itsPlotMenu, kFitParmsCmd, menuBar);
	assert( itsFitParmsMenu != NULL );
	itsFitParmsMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsFitParmsMenu);
	itsPlotMenu->DisableItem(kFitParmsCmd);
	
	itsDiffMenu = new JXTextMenu(itsPlotMenu, kDiffPlotCmd, menuBar);
	assert( itsDiffMenu != NULL );
	itsDiffMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsDiffMenu);
	itsPlotMenu->DisableItem(kDiffPlotCmd);

	
	itsCurveMenu = new JXTextMenu(itsPlotMenu, kLinearFitCmd, menuBar);
	assert( itsCurveMenu != NULL );
	itsCurveMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsCurveMenu);
	
	itsCurveExpMenu = new JXTextMenu(itsPlotMenu, kExpFitCmd, menuBar);
	assert( itsCurveExpMenu != NULL );
	itsCurveExpMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsCurveExpMenu);
	
	itsCurveQuadMenu = new JXTextMenu(itsPlotMenu, kQuadFitCmd, menuBar);
	assert( itsCurveQuadMenu != NULL );
	itsCurveQuadMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsCurveQuadMenu);

	itsCurveModuleMenu = new JXTextMenu(itsPlotMenu, kFitModuleCmd, menuBar);
	assert( itsCurveModuleMenu != NULL );
	itsCurveModuleMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsCurveModuleMenu);

	itsSessionDir = 
		new GloveHistoryDir(JXGetApplication());
	assert(itsSessionDir != NULL);
	JXGetDocumentManager()->DocumentMustStayOpen(itsSessionDir, kTrue);
	ListenTo(itsSessionDir);
	
	itsPlot = 
		new GlovePlotter(itsSessionDir, menuBar, window, 
			JXWidget::kHElastic, JXWidget::kVElastic, 
			0, 30, w, h - kJXStdMenuBarHeight);
	assert( itsPlot != NULL );

	ListenTo(itsPlot);

	JXDocumentMenu* windowListMenu = 
		new JXDocumentMenu(kWindowListMenuTitleStr, menuBar,
			JXWidget::kFixedLeft, JXWidget::kVElastic, 0,0, 10,10);
	assert( windowListMenu != NULL );
	menuBar->AppendMenu(windowListMenu);

	itsHelpMenu = menuBar->AppendTextMenu(kHelpMenuTitleStr);
	itsHelpMenu->SetMenuItems(kHelpMenuStr);
	itsHelpMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsHelpMenu);
	
	itsVarList	 = new GVarList();
	itsVarList->AddVariable("x",0);
	itsXVarIndex = 1;
	itsFunctionDialog = NULL;
	itsFitModuleDialog = NULL;
	
	itsFits = new JPtrArray<JFitBase>;
	assert( itsFits != NULL );
	ListenTo(itsFits);
	itsFitParmsDir = new FitParmsDir(this, itsFits);
	assert( itsFitParmsDir != NULL );
	itsCurveStats = new JArray<CurveStats>;
	assert( itsCurveStats != NULL );
	itsCurveIndeces = new JArray<JIndex>;
	assert( itsCurveIndeces != NULL );
	itsCurrentCurveType = kGDataCurve;
	
	itsDiffDirs = new JPtrArray<PlotDir>;
	assert( itsDiffDirs != NULL );
	ListenTo(itsDiffDirs);
	
	itsPrinter = new JXPSPrinter(GetDisplay(), (GetWindow())->GetColormap());
	assert( itsPrinter != NULL );
	itsPrinter->SetOrientation(JPSPrinter::kLandscape);

	ListenTo(itsPrinter);
}

/******************************************************************************
 Destructor

 ******************************************************************************/

PlotDir::~PlotDir()
{
	delete itsFits;
	delete itsCurveStats;
//	for (JSize i = itsDiffDirs->GetElementCount(); i >= 1; i++)
//		{
//		itsDiffDirs->NthElement(i)->Close();
//		}
	delete itsDiffDirs;
	itsSessionDir->Close();
}

/******************************************************************************
 Receive

 ******************************************************************************/

void
PlotDir::Receive
	(
	JBroadcaster* sender,
	const JBroadcaster::Message& message
	)
{
	if (sender == itsPlot && message.Is(J2DPlotWidget::kTitleChanged))
		{
		JString title = *itsFileName + ":  " + itsPlot->GetTitle();
		(GetWindow())->SetTitle(title);
		JString sessiontitle = "Glove session  -  " + title;
		(itsSessionDir->GetWindow())->SetTitle(sessiontitle);
		}
		
	else if (sender == itsPlot && message.Is(J2DPlotWidget::kPlotChanged))
		{
		itsSupervisor->DataModified();
		UpdateCurveMenu();
		}

	else if (sender == itsSessionDir && message.Is(GloveHistoryDir::kSessionChanged))
		{
		itsSupervisor->DataModified();
		}
		
	else if (sender == itsPlot && message.Is(J2DPlotWidget::kIsEmpty))
		{
		if (!itsPlotIsClosing)
			{
			GLClosePlotDirTask* task = new GLClosePlotDirTask(this);
			assert(task != NULL);
			JXGetApplication()->InstallUrgentTask(task);
			}
		}
		
	else if (sender == itsPlot && message.Is(J2DPlotWidget::kCurveAdded))
		{
		HandleCurveAdded();
		UpdateCurveMenu();
		}
		
	else if (sender == itsPlot && message.Is(J2DPlotWidget::kCurveRemoved))
		{
		const J2DPlotWidget::CurveRemoved* info =
			dynamic_cast(const J2DPlotWidget::CurveRemoved*, &message);
		assert( info != NULL );
		HandleCurveRemoved(info->GetIndex());
		UpdateCurveMenu();
		}
		
	else if (sender == itsPlotMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandlePlotMenu(selection->GetIndex());
		}
	else if (sender == itsHelpMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleHelpMenu(selection->GetIndex());
		}
	
	else if (sender == itsFits && 
		( 	message.Is(JOrderedSetT::kElementsInserted) ||
			message.Is(JOrderedSetT::kElementsRemoved) ||
			message.Is(JOrderedSetT::kAllElementsRemoved)))
		{
		UpdateFitParmsMenu();
		}	
		
	else if (sender == itsDiffDirs && 
		( 	message.Is(JOrderedSetT::kElementsInserted) ||
			message.Is(JOrderedSetT::kElementsRemoved) ||
			message.Is(JOrderedSetT::kAllElementsRemoved)))
		{
		UpdateDiffMenu();
		}	

		
	else if (sender == itsFitParmsMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		itsFitParmsDir->Activate();
		itsFitParmsDir->ShowFit(selection->GetIndex());
		}
		
	else if (sender == itsDiffMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		PlotDir* dir = itsDiffDirs->NthElement(selection->GetIndex());
		dir->Activate();
//		(itsDiffDirs->NthElement(selection->GetIndex()))->Activate();
		}

		
	else if (sender == itsCurveMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleCurveMenu(selection->GetIndex());
		}
		
	else if (sender == itsCurveExpMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleCurveExpMenu(selection->GetIndex());
		}

	else if (sender == itsCurveQuadMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleCurveQuadMenu(selection->GetIndex());
		}
		
	else if (sender == itsCurveModuleMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleCurveModuleMenu(selection->GetIndex());
		}	
		
	else if (sender == itsFunctionDialog && message.Is(JXDialogDirector::kDeactivated))
		{
		const JXDialogDirector::Deactivated* info =
			dynamic_cast(const JXDialogDirector::Deactivated*, &message);
		assert( info != NULL );
		if (info->Successful())
			{
			JString fnStr = itsFunctionDialog->GetFunctionString();
			JFunction* f;
			if (JParseFunction(fnStr, itsVarList, &f))
				{
				PlotFunction(f);
				}
			}
		itsFunctionDialog = NULL;
		}
		
	else if (sender == itsFitModuleDialog && message.Is(JXDialogDirector::kDeactivated))
		{
		const JXDialogDirector::Deactivated* info =
			dynamic_cast(const JXDialogDirector::Deactivated*, &message);
		assert( info != NULL );
		if (info->Successful())
			{
			JIndex index = itsFitModuleDialog->GetFilterIndex();
			JPlotDataBase* data = &(itsPlot->GetCurve(itsCurveForFit));
			FitModule* fm;
			JString modName;
			gPlotApp->GetFitModulePath(index, &modName);
			JBoolean success = FitModule::Create(&fm, this, data, modName);
			if (!success)
				{
				JGetUserNotification()->ReportError("Unknown module error.");
				}
			}
		itsFitModuleDialog = NULL;
		}
	
	else if (sender == itsPrinter &&
			 message.Is(JPrinter::kPageSetupFinished))
		{
		const JPrinter::PageSetupFinished* info = 
			dynamic_cast(const JPrinter::PageSetupFinished*, &message);
		assert(info != NULL);
		if (info->Changed())
			{
			itsSupervisor->DataModified();
			// for WYSISYG: adjust display to match new paper size
			}
		}
			
	else if (sender == itsPrinter &&
			 message.Is(JPrinter::kPrintSetupFinished))
		{
		const JPrinter::PrintSetupFinished* info = 
			dynamic_cast(const JPrinter::PrintSetupFinished*, &message);
		assert(info != NULL);
		if (info->Successful())
			{
			itsPlot->Print(*itsPrinter);
			if (itsIsPrintAll)
				{
				itsFitParmsDir->SendAllToSession();
				itsSessionDir->Print();
				}
			}
		}
	else
		{
		JXWindowDirector::Receive(sender, message);
		}
}

/******************************************************************************
 NewFileName

 ******************************************************************************/

void
PlotDir::NewFileName
	(
	const JCharacter* filename
	)
{
	*itsFileName = filename;
	JString title = *itsFileName + ":  " + itsPlot->GetTitle();
	(GetWindow())->SetTitle(title);
	JString sessiontitle = "Glove session  -  " + title;
	(itsSessionDir->GetWindow())->SetTitle(sessiontitle);
}

/******************************************************************************
 WriteSetup

 ******************************************************************************/

void
PlotDir::WriteSetup
	(
	ostream& os
	)
{
	GetWindow()->WriteGeometry(os);
	itsPrinter->WriteXPSSetup(os);
}

/******************************************************************************
 ReadSetup

 ******************************************************************************/

void
PlotDir::ReadSetup
	(
	istream& is
	)
{
	GetWindow()->ReadGeometry(is);
	itsPrinter->ReadXPSSetup(is);
}

/******************************************************************************
 WriteData

 ******************************************************************************/

void
PlotDir::WriteData
	(
	ostream& os,
	GRaggedFloatTableData* data
	)
{
	os << kCurrentSetupVersion << ' ';


	const JSize tempCount = itsPlot->GetCurveCount();
	JSize count = 0;
	JSize i;
	for (i = 1; i <= tempCount; i++)
		{
		JPlotDataBase& jpdb = itsPlot->GetCurve(i);
		CurveStats stat = itsCurveStats->GetElement(i);
		if ((stat.type == kGDataCurve) && (jpdb.GetType() == JPlotDataBase::kScatterPlot))
			{
			J2DPlotData* pd = dynamic_cast(J2DPlotData*, &jpdb);
			if (pd->IsValid())
				{
				count++;
				}
			}
		else if (stat.type == kGFitCurve)
			{
			JPlotDataBase& pdb = itsPlot->GetCurve(stat.provider);
			J2DPlotData* pd = dynamic_cast(J2DPlotData*, &pdb);
			if (pd->IsValid())
				{
				count++;
				}
			}
		else
			{
			count++;
			}
		}
	os << count << ' ';
	for (i = 1; i <= tempCount; i++)
		{
		JPlotDataBase& jpdb = itsPlot->GetCurve(i);
		CurveStats stat = itsCurveStats->GetElement(i);
		if (stat.type == kGDataCurve)
			{
			if (jpdb.GetType() == JPlotDataBase::kScatterPlot)
				{
				J2DPlotData* pd = dynamic_cast(J2DPlotData*, &jpdb);
				if (pd->IsValid())
					{
					os << (int)kGDataCurve << ' ';
					os << (int)jpdb.GetType() << ' ';
					JIndex index;
					data->FindColumn(pd->GetXData(), &index);
					os << index << ' ';
					data->FindColumn(pd->GetYData(), &index);
					os << index << ' ';
					os << pd->HasXErrors() << ' ';
					os << pd->HasYErrors() << ' ';
					if (pd->HasXErrors())
						{
						const JArray<JFloat>* testArray;
						pd->GetXErrorData(&testArray);
						data->FindColumn(testArray, &index);
						os << index << ' ';
						}
					if (pd->HasYErrors())
						{
						const JArray<JFloat>* testArray;
						pd->GetYErrorData(&testArray);
						data->FindColumn(testArray, &index);
						os << index << ' ';
						}
					}
				}
			else if (jpdb.GetType() == JPlotDataBase::kVectorPlot)
				{
				J2DVectorData* vd = dynamic_cast(J2DVectorData*, &jpdb);
				os << (int)kGDataCurve << ' ';
				os << (int)jpdb.GetType() << ' ';
				JIndex index;
				data->FindColumn(vd->GetXData(), &index);
				os << index << ' ';
				data->FindColumn(vd->GetYData(), &index);
				os << index << ' ';
				data->FindColumn(vd->GetVXData(), &index);
				os << index << ' ';
				data->FindColumn(vd->GetVYData(), &index);
				os << index << ' ';
				}
			}
		else if (stat.type == kGFitCurve)
			{
			if (stat.fitType == kGModFit)
				{
				os << (int)kGFitCurve << ' ';
				os << (int)stat.fitType << ' ';
				os << stat.provider << ' ';
				JPlotModuleFit* mf = dynamic_cast(JPlotModuleFit*, &jpdb);
				mf->WriteData(os);
				}
			else
				{
				JPlotDataBase& pdb = itsPlot->GetCurve(stat.provider);
				J2DPlotData* pd = dynamic_cast(J2DPlotData*, &pdb);
				if (pd->IsValid())
					{
					os << (int)kGFitCurve << ' ';
					os << (int)stat.fitType << ' ';
					os << stat.provider << ' ';
					}
				}
			}
		else if (stat.type == kGFunctionCurve)
			{
			JPlotJFunction* pd = dynamic_cast(JPlotJFunction*, &jpdb);
			os << (int)kGFunctionCurve << ' ';
			os << pd->GetFunctionString() << ' ';
			}
		else
			{
			os << (int)kGDiffCurve << ' ';
			}
		}
	itsPlot->PWWriteSetup(os);
	itsPlot->PWWriteCurveSetup(os);
	const JSize diffCount = itsDiffDirs->GetElementCount();
	os << ' ' << diffCount << ' ';
	for (i = 1; i <= diffCount; i++)
		{
		(itsDiffDirs->NthElement(i))->WriteSetup(os);
		(itsDiffDirs->NthElement(i))->WriteData(os, data);
		}
	itsSessionDir->WriteData(os);
}

/******************************************************************************
 ReadData

 ******************************************************************************/

void
PlotDir::ReadData
	(
	istream& is,
	GRaggedFloatTableData* data
	)
{
	JFileVersion version = 0;
	is >> ws;
	is >> version;

	if (version == kCurrentSetupVersion)
		{
		ReadCurves(is,data);
		itsPlot->PWReadSetup(is);
		}
	else if (version == 0)
		{
		itsPlot->PWReadSetup(is);
		ReadCurves(is,data);
		}
	if (version <= kCurrentSetupVersion)
		{
		itsPlot->PWReadCurveSetup(is);
		JSize diffCount;
		is >> diffCount;
		assert(diffCount == itsDiffDirs->GetElementCount());
		JSize i;
		for (i = 1; i <= diffCount; i++)
			{
			(itsDiffDirs->NthElement(i))->ReadSetup(is);
			(itsDiffDirs->NthElement(i))->ReadData(is, data);
			}
		itsSessionDir->ReadData(is);
		}
}

/******************************************************************************
 ReadCurves (private)
 
 ******************************************************************************/

void
PlotDir::ReadCurves
	(
	istream& is,
	GRaggedFloatTableData* data
	)
{
	JSize curveCount;
	is >> curveCount;

	for (JSize j = 1; j <= curveCount; j++)
		{
		GCurveType type;
		int temp;
		is >> temp;
		type = (GCurveType)temp;
		if (type == kGDataCurve)
			{
			is >> temp;
			if (temp == JPlotDataBase::kScatterPlot)
				{		
				JIndex xIndex;
				is >> xIndex;
				JIndex yIndex;
				is >> yIndex;
				JArray<JFloat>* xCol = data->GetColPointer(xIndex);
				JArray<JFloat>* yCol = data->GetColPointer(yIndex);
				JArray<JFloat>* xErrCol;
				JArray<JFloat>* yErrCol;
				JBoolean xerrors;
				is >> xerrors;
				JBoolean yerrors;
				is >> yerrors;
				if (xerrors)
					{
					is >> xIndex;
					xErrCol = data->GetColPointer(xIndex);
					}
				if (yerrors)
					{
					is >> yIndex;
					yErrCol = data->GetColPointer(yIndex);
					}
				itsCurrentCurveType = kGDataCurve;
				J2DPlotData* curve;
				if (J2DPlotData::Create(&curve, *xCol, *yCol, kTrue))
					{
					itsPlot->AddCurve(curve, kTrue, "");
					}
				else
					{
					return;
					}
				if (xerrors)
					{
					curve->SetXErrors(*xErrCol);
					}
				if (yerrors)
					{
					curve->SetYErrors(*yErrCol);
					}
				}
			else if (temp == JPlotDataBase::kVectorPlot)
				{
				JIndex index;
				is >> index;
				JArray<JFloat>* xCol = data->GetColPointer(index);
				is >> index;
				JArray<JFloat>* yCol = data->GetColPointer(index);
				is >> index;
				JArray<JFloat>* vxCol = data->GetColPointer(index);
				is >> index;
				JArray<JFloat>* vyCol = data->GetColPointer(index);
				itsCurrentCurveType = kGDataCurve;
				J2DVectorData* curve;
				if (J2DVectorData::Create(&curve, *xCol, *yCol, *vxCol, *vyCol, kTrue))
					{
					itsPlot->AddCurve(curve, kTrue, "", kFalse, kFalse, kTrue);
					}
				else
					{
					return;
					}

				}
			}
		else if (type == kGFitCurve)
			{
			GCurveFitType ftype;
			int temp;
			is >> temp;
			ftype = (GCurveFitType)temp;
			JIndex provider;
			is >> provider;
			if (ftype == kGModFit)
				{
				JPlotDataBase* pdata = &(itsPlot->GetCurve(provider));
				JPlotModuleFit* mf = new JPlotModuleFit(itsPlot, pdata, is);
				assert(mf != NULL);
				if (!AddFitModule(mf, pdata))
					{
					delete mf;
					}
				}
			else
				{
				NewFit(provider, ftype);
				}
			}
		else if (type == kGFunctionCurve)
			{
			JString fnString;
			is >> fnString;
			JPlotJFunction* pf;
			JFloat min, max, inc;
			itsPlot->GetXScaling(&min, &max, &inc);
			if (JPlotJFunction::Create(&pf, itsPlot, itsVarList, fnString, itsXVarIndex, min, max))
				{
				itsCurrentCurveType = kGFunctionCurve;
				itsPlot->AddCurve(pf, kTrue, pf->GetFunctionString(), kTrue, kFalse);
				itsCurrentCurveType = kGDataCurve;
				}
			}
		}
}

/******************************************************************************
 OKToClose

 ******************************************************************************/

JBoolean
PlotDir::OKToClose()
{
	return kTrue;	
}

/******************************************************************************
 NeedsSave

 ******************************************************************************/

JBoolean
PlotDir::NeedsSave()
	const
{
	return kFalse;
}

/******************************************************************************
 HandlePlotMenu

 ******************************************************************************/

void
PlotDir::HandlePlotMenu
	(
	const JIndex index
	)
{
	if (index == kPageSetupCmd)
		{
		itsPrinter->BeginUserPageSetup();
		}
	else if (index == kPrintCmd)
		{
		itsPrinter->BeginUserPrintSetup();
		itsIsPrintAll = kFalse;
		}
	else if (index == kPrintSessionCmd)
		{
		itsPrinter->BeginUserPrintSetup();
		itsIsPrintAll = kTrue;
		}
	else if (index == kPlotFunctionCmd)
		{
		CreateFunction();
		}
	else if (index == kCloseCmd)
		{
		if (itsHideOnClose)
			{
			Deactivate();
			}
		else
			{
			Close();
			}
		}
}

/******************************************************************************
 PlotFunction

 ******************************************************************************/

void
PlotDir::PlotFunction
	(
	JFunction* f
	)
{
	JFloat min, max, inc;
	itsPlot->GetXScaling(&min, &max, &inc);
	JPlotJFunction* pf = new JPlotJFunction(itsPlot, itsVarList, f, kTrue, itsXVarIndex, min, max);
	itsCurrentCurveType = kGFunctionCurve;
	itsPlot->AddCurve(pf, kTrue, f->Print(), kTrue, kFalse);
	itsCurrentCurveType = kGDataCurve;	
}

/******************************************************************************
 CreateFunction

 ******************************************************************************/

void
PlotDir::CreateFunction()
{
	assert(itsFunctionDialog == NULL);
	itsFunctionDialog = 
		new GXPlotFunctionDialog(this, itsVarList);
	assert( itsFunctionDialog != NULL);
	ListenTo(itsFunctionDialog);
	itsFunctionDialog->BeginDialog();
}

/******************************************************************************
 UpdateFitParmsMenu

 ******************************************************************************/

void
PlotDir::UpdateFitParmsMenu()
{
	itsFitParmsMenu->RemoveAllItems();
	const JSize count = itsFits->GetElementCount();
	JSize i;
	for (i = 1; i <= count; i++)
		{
		JString numS(i);
		JString str = "Fit " + numS;
		itsFitParmsMenu->AppendItem(str);
		}
	if (count == 0)
		{
		itsPlotMenu->DisableItem(kFitParmsCmd);
		}

}

/******************************************************************************
 UpdateDiffMenu

 ******************************************************************************/

void
PlotDir::UpdateDiffMenu()
{
	itsDiffMenu->RemoveAllItems();
	const JSize count = itsDiffDirs->GetElementCount();
	JSize i;
	for (i = 1; i <= count; i++)
		{
		JString numS(i);
		JString str = "Residual plot " + numS;
		itsDiffMenu->AppendItem(str);
		}
	if (count == 0)
		{
		itsPlotMenu->DisableItem(kDiffPlotCmd);
		}

}

/******************************************************************************
 UpdateCurveMenu

 ******************************************************************************/

void
PlotDir::UpdateCurveMenu()
{
	itsCurveMenu->RemoveAllItems();
	itsCurveExpMenu->RemoveAllItems();
	itsCurveQuadMenu->RemoveAllItems();
	itsCurveModuleMenu->RemoveAllItems();

	itsCurveIndeces->RemoveAll();
	JSize i;
	for (i = 1; i <= itsPlot->GetCurveCount(); i++)
		{
		CurveStats stats = itsCurveStats->GetElement(i);
		if ((stats.type == kGDataCurve) || (stats.type == kGDiffCurve))
			{
			itsCurveMenu->AppendItem(itsPlot->GetCurveName(i));
			itsCurveExpMenu->AppendItem(itsPlot->GetCurveName(i));
			itsCurveQuadMenu->AppendItem(itsPlot->GetCurveName(i));
			itsCurveModuleMenu->AppendItem(itsPlot->GetCurveName(i));
			itsCurveIndeces->AppendElement(i);
			}
		}
}

/******************************************************************************
 HandleCurveMenu

 ******************************************************************************/

void
PlotDir::HandleCurveMenu
	(
	const JIndex index
	)
{
	JIndex plotIndex = itsCurveIndeces->GetElement(index);
	NewFit(plotIndex, kGLinearFit);
}

/******************************************************************************
 HandleCurveExpMenu

 ******************************************************************************/

void
PlotDir::HandleCurveExpMenu
	(
	const JIndex index
	)
{
	JIndex plotIndex = itsCurveIndeces->GetElement(index);
	NewFit(plotIndex, kGExpFit);
}

/******************************************************************************
 HandleCurveQuadMenu

 ******************************************************************************/

void
PlotDir::HandleCurveQuadMenu
	(
	const JIndex index
	)
{
	JIndex plotIndex = itsCurveIndeces->GetElement(index);
	NewFit(plotIndex, kGQuadFit);
}

/******************************************************************************
 HandleCurveModuleMenu

 ******************************************************************************/

void
PlotDir::HandleCurveModuleMenu
	(
	const JIndex index
	)
{
	JIndex plotIndex = itsCurveIndeces->GetElement(index);
//	JPlotDataBase* data = &(itsPlot->GetCurve(plotindex));
	itsCurveForFit = plotIndex;
	SelectFitModule();
//	FitModule* fm;
//	if (FitModule::Create(&fm, this, data

}

/******************************************************************************
 SelectFitModule

 ******************************************************************************/

void
PlotDir::SelectFitModule()
{
	assert(itsFitModuleDialog == NULL);
	itsFitModuleDialog = 
		new FitModuleDialog(this);
	assert( itsFitModuleDialog != NULL);
	ListenTo(itsFitModuleDialog);
	itsFitModuleDialog->BeginDialog();
}

/******************************************************************************
 HandleCurveAdded

 ******************************************************************************/

void
PlotDir::HandleCurveAdded()
{
	CurveStats stats;
	stats.type = itsCurrentCurveType;
	stats.provider = 0;
	stats.fitNumber = 0;
	itsCurveStats->AppendElement(stats);
}

/******************************************************************************
 HandleCurveRemoved

 ******************************************************************************/

void
PlotDir::HandleCurveRemoved
	(
	const JIndex index
	)
{
	CurveStats stats = itsCurveStats->GetElement(index);
	if (stats.type == kGFitCurve)
		{
		RemoveFit(index);
		}
	itsCurveStats->RemoveElement(index);
	const JSize ccount = itsCurveStats->GetElementCount();
	JSize count = ccount;
	JSize i;
	if (stats.type == kGDataCurve)
		{
		i = index;
		while (i <= count)
			{
			CurveStats statsi = itsCurveStats->GetElement(i);
			if (statsi.type == kGFitCurve)
				{
				if (statsi.provider == index)
					{
					itsPlot->RemoveCurve(i);
					count --;
					}
				else
					{
					statsi.provider--;
					itsCurveStats->SetElement(i, statsi);
					i++;
					}
				}
			else
				{
				i++;
				}
			}
		}
	else if (stats.type == kGFunctionCurve)
		{
		for (i = index + 1; i <= count; i++)
			{
			CurveStats statsi = itsCurveStats->GetElement(i);
			if (statsi.type == kGFitCurve)
				{
				statsi.provider--;
				itsCurveStats->SetElement(i, statsi);
				}
			}
		}
}

/******************************************************************************
 RemoveFit

 ******************************************************************************/

void
PlotDir::RemoveFit
	(
	const JIndex index
	)
{
	CurveStats stats = itsCurveStats->GetElement(index);
	itsFits->RemoveElement(stats.fitNumber);
	PlotDir* dir = itsDiffDirs->NthElement(stats.fitNumber);
	dir->Close();
	itsDiffDirs->RemoveElement(stats.fitNumber);
	const JSize count = itsCurveStats->GetElementCount();
	JSize i;
	for (i = index + 1; i <= count; i++)
		{
		stats = itsCurveStats->GetElement(i);
		if (stats.type == kGFitCurve)
			{
			stats.provider--;
			stats.fitNumber--;
			itsCurveStats->SetElement(i, stats);
			}
		}
}

/******************************************************************************
 NewFit

 ******************************************************************************/

void
PlotDir::NewFit
	(
	const JIndex plotindex,
	const GCurveFitType type
	)
{
	JPlotDataBase* data = &(itsPlot->GetCurve(plotindex));
	JPlotFitFunction* df;
	if (type == kGLinearFit)
		{
		JPlotLinearFit* lf;
		JFloat xmax, xmin, ymax, ymin;
		if (itsPlot->IsUsingRange())
			{
			itsPlot->GetRange(&xmin, &xmax, &ymin, &ymax);
			lf = new JPlotLinearFit(itsPlot, data, xmin, xmax, ymin, ymax);
			}
		else
			{
			data->GetXRange(&xmin, &xmax);
			lf = new JPlotLinearFit(itsPlot, data, xmin, xmax);
			}
		assert(lf != NULL);
		itsFits->Append(lf);
		df = lf;
		}
	else if (type == kGExpFit)
		{
		JPlotLinearFit* lf;
		JFloat xmax, xmin, ymax, ymin;
		itsPlot->GetRange(&xmin, &xmax, &ymin, &ymax);
		if (itsPlot->IsUsingRange())
			{
			itsPlot->GetRange(&xmin, &xmax, &ymin, &ymax);
			lf = new JPlotLinearFit(itsPlot, data, xmin, xmax, ymin, ymax, kFalse, kTrue);
			}
		else
			{
			data->GetXRange(&xmin, &xmax);
			lf = new JPlotLinearFit(itsPlot, data, xmin, xmax, kFalse, kTrue);
			}
		assert(lf != NULL);
		itsFits->Append(lf);
		df = lf;
//		JPlotExpFit* ef = new JPlotExpFit(itsPlot, data);
//		assert(ef != NULL);
//		itsFits->Append(ef);
//		df = ef;
		}
	else if (type == kGQuadFit)
		{
		JPlotQuadFit* lf;
		JFloat xmax, xmin, ymax, ymin;
		if (itsPlot->IsUsingRange())
			{
			itsPlot->GetRange(&xmin, &xmax, &ymin, &ymax);
			lf = new JPlotQuadFit(itsPlot, data, xmin, xmax, ymin, ymax);
			}
		else
			{
			data->GetXRange(&xmin, &xmax);
			lf = new JPlotQuadFit(itsPlot, data, xmin, xmax);
			}
		assert(lf != NULL);
		itsFits->Append(lf);
		df = lf;
		}
	AddFit(df, plotindex, type);

}

/******************************************************************************
 AddDiffCurve

 ******************************************************************************/

void
PlotDir::AddDiffCurve
	(
	JPlotDataBase* ddata
	)
{
	itsCurrentCurveType = kGDiffCurve;
	itsPlot->AddCurve(ddata, kFalse, "Residual data", kFalse, kTrue);
	itsCurrentCurveType = kGDataCurve;
}

/******************************************************************************
 AddFitModule

 ******************************************************************************/

JBoolean
PlotDir::AddFitModule
	(
	JPlotModuleFit* fit,
	JPlotDataBase* fitData
	)
{
	JIndex plotindex;
	JBoolean found = itsPlot->GetCurveIndex(fitData, &plotindex);
	if (!found)
		{
		return kFalse;
		}
	itsFits->Append(fit);
	AddFit(fit, plotindex, kGModFit);
	return kTrue;
}
		
/******************************************************************************
 AddFit

 ******************************************************************************/

void
PlotDir::AddFit
	(
	JPlotFitFunction* fit,
	const JIndex plotindex,
	const GCurveFitType type
	)
{
	itsCurrentCurveType = kGFitCurve;
	JIndex i = itsPlot->AddCurve(fit, kTrue, fit->GetFunctionString(), kTrue, kFalse);
	itsCurrentCurveType = kGDataCurve;
	CurveStats stats = itsCurveStats->GetElement(i);
	stats.type 		= kGFitCurve;
	stats.provider	= plotindex;
	stats.fitNumber = itsFits->GetElementCount();
	stats.fitType	= type; 
	itsCurveStats->SetElement(i, stats);
	itsPlotMenu->EnableItem(kFitParmsCmd);
	itsPlotMenu->EnableItem(kDiffPlotCmd);
	
	PlotDir* dir = new PlotDir(this, itsSupervisor, *itsFileName, kTrue);
	assert( dir != NULL );
	JXGetDocumentManager()->DocumentMustStayOpen(dir, kTrue);
	JPlotDataBase* ddata = fit->GetDiffData();
	dir->AddDiffCurve(ddata);
	J2DPlotWidget* plot = dir->GetPlot();
	JString numS(itsFits->GetElementCount());
	JString str = "Residual Plot " + numS;
	plot->SetTitle(str);
	plot->SetXLabel(itsPlot->GetXLabel());
	plot->SetYLabel(itsPlot->GetYLabel());
	plot->ProtectCurve(1, kTrue);
	plot->ShowFrame(kFalse);
	itsDiffDirs->Append(dir);
}

/******************************************************************************
 HandleHelpMenu

 ******************************************************************************/

void
PlotDir::HandleHelpMenu
	(
	const JIndex index
	)
{
	if (index == kAboutCmd)
		{
		GetGLApplication()->DisplayAbout();
		}
	else if (index == kTOCCmd)
		{
		JXGetHelpManager()->ShowTOC();
		}
	else if (index == kThisWindowCmd)
		{
		JXGetHelpManager()->ShowSection(kPlotHelpName);
		}
	else if (index == kChangesCmd)
		{
		JXGetHelpManager()->ShowSection(kChangeLogName);
		}
	else if (index == kCreditsCmd)
		{
		JXGetHelpManager()->ShowSection(kCreditsName);
		}
}

/******************************************************************************
 SafetySave

 ******************************************************************************/

void
PlotDir::SafetySave
	(
	const JXDocumentManager::SafetySaveReason reason
	)
{

}

/******************************************************************************
 DiscardChanges

 ******************************************************************************/

void
PlotDir::DiscardChanges()
{

}

/******************************************************************************
 GetMenuIcon (virtual)

	Override of JXDocument::GetMenuIcon().

 ******************************************************************************/

JXPM
PlotDir::GetMenuIcon()
	const
{
	return JXGetPlainFileIcon();
}

/******************************************************************************
 Close (public)

 ******************************************************************************/

JBoolean
PlotDir::Close()
{
	itsPlotIsClosing = kTrue;
	return JXDocument::Close();
}
