/******************************************************************************
 TestFloatTableDirector.cc

	BASE CLASS = JXWindowDirector

	Written by John Lindal.

 ******************************************************************************/

#include "TestFloatTableDirector.h"
#include "GXRaggedFloatTable.h"
#include "GloveModule.h"
#include "GlovePlotter.h"
#include "GXPlotApp.h"
#include "GXChooseFileImportDialog.h"
#include "GXGetDelimiterDialog.h"
#include "PlotDir.h"
#include "JX2DPlotWidget.h"
#include "J2DPlotData.h"
#include "J2DVectorData.h"
#include "glovedatadoc.h"
#include <GXRowHeaderWidget.h>
#include <GXColHeaderWidget.h>
#include <GRaggedFloatTableData.h>
#include <GLGlobals.h>
#include "GLHelpText.h"
#include "GXPlotApp.h"

#include <filenew.xpm>
#include <fileopen.xpm>
#include <filefloppy.xpm>
#include <fileprint.xpm>
#include <manual.xpm>

#include <JXToolBar.h>

#include <JXDialogDirector.h>
#include <JXHelpDirector.h>
#include <JXDocumentMenu.h>
#include <JXWindow.h>
#include <JXMenuBar.h>
#include <JXTextMenu.h>
#include <JXScrollbarSet.h>
#include <JXScrollbar.h>
#include <JXApplication.h>
#include <JXPSPrinter.h>
#include <JXImage.h>
#include <JXTextButton.h>

#include <JArray.h>
#include <JPtrArray.h>
#include <JCreateProgressDisplay.h>
#include <JUserNotification.h>
#include <JChooseSaveFile.h>
#include <JString.h>
#include <JInPipeStream.h>
#include <JOutPipeStream.h>

#include <jErrno.h>
#include <jProcessUtil.h>
#include <jStrStreamUtil.h>
#include <jStreamUtil.h>

#include <iostream.h>
#include <fstream.h>
#include <jAssert.h>

extern GXPlotApp* gPlotApp;

// File menu information

static const JCharacter* kFileMenuTitleStr  = "File";
static const JCharacter* kFileMenuStr =
	"New %k Meta-N | Open... %k Meta-O | Save %k Meta-S | Save as..."
	"%l|Export modules..."
	"%l| Page setup... | Print... %k Meta-P" 
	"%l| Close %k Meta-W | Quit %k Meta-Q";

static const JCharacter* kGloveFileSignature = "*** Glove File Format ***";
static const JFloat kCurrentGloveVersion = 1.0;

enum
{
	kNewCmd = 1,
	kOpenCmd,
	kSaveCmd,
	kSaveAsCmd,
	kExportCmd,
	kPageSetupCmd,
	kPrintCmd,
	kCloseCmd,
	kQuitCmd
};

const JSize kMaxPlotTitleSize = 20;

static const JCharacter* kWindowListMenuTitleStr = "Windows";

static const JCharacter* kExportMenuTitleStr = "Export modules";
static const JCharacter* kExportMenuStr =
	"Reload %l";

enum
{
	kReloadModuleCmd = 1
};

static const JCharacter* kHelpMenuTitleStr = "Help";
static const JCharacter* kHelpMenuStr = 
	"About %l| Table of Contents | This window "
	"%l|Changes|Credits";

enum
{
	kAboutCmd = 1,
	kTOCCmd,
	kThisWindowCmd,
	kChangesCmd,
	kCreditsCmd
};

enum
{
	kDataPlot = 1,
	kVectorPlot
};

const JCoordinate kButtConBuffer 	= 5;
const JCoordinate kButtConWidth 	= 30;
const JCoordinate kToolBarHeight 	= 40;

const JCoordinate kEditButtonStart	= 130;

static const JCharacter* kInternalModuleNames[] =
{
	"Text (delimited)",
	"Text (fixed-width)"
};

const JSize kInternalModuleCount = sizeof(kInternalModuleNames)/sizeof(JCharacter*);

enum
{
	kDelimitedText = 1,
	kFixedWidthText
};

/******************************************************************************
 Constructor

 ******************************************************************************/

TestFloatTableDirector::TestFloatTableDirector
	(
	JXDirector* supervisor,
	const JCharacter* fileName,
	const JBoolean onDisk
	)
	:
	JXFileDocument(supervisor, fileName, onDisk, kFalse, ".glv"),
	itsDelimiterDialog(NULL)
{
	itsData = new GRaggedFloatTableData(0.0);
	assert( itsData != NULL );
	
	itsPrinter          = NULL;

	BuildWindow();

	itsPrinter = new JXPSPrinter(GetDisplay(), (GetWindow())->GetColormap());
	assert( itsPrinter != NULL );
	ListenTo(itsPrinter);
	
	itsFileImportDialog = NULL;
	
	itsPlotWindows = new JPtrArray<PlotDir>();
	
	itsPlotNumber = 1;
	AdjustWindowTitle();
	UpdateExportMenu();
	itsListenToData = kTrue;
	ListenTo(itsData);
}

/******************************************************************************
 Destructor

 ******************************************************************************/

TestFloatTableDirector::~TestFloatTableDirector()
{
	delete itsData;
	delete itsPrinter;
	delete itsPlotWindows;
}

/******************************************************************************
 BuildWindow (private)

 ******************************************************************************/

void
TestFloatTableDirector::BuildWindow()
{
	JSize w = 453;
	JSize h = 360;

    JXWindow* window = new JXWindow(this, w,h, "");
    assert( window != NULL );
    SetWindow(window);

    JXMenuBar* menuBar =
        new JXMenuBar(window,
                    JXWidget::kHElastic, JXWidget::kFixedTop, 
                    0,0, w,kJXStdMenuBarHeight);
    assert( menuBar != NULL );

    JXToolBar* toolBar = 
    	new JXToolBar(window,
                    JXWidget::kHElastic, JXWidget::kFixedTop, 
                    0,kJXStdMenuBarHeight, 
                    w,kToolBarHeight);
    assert( toolBar != NULL );

	itsFileMenu = menuBar->AppendTextMenu(kFileMenuTitleStr);
	itsFileMenu->SetMenuItems(kFileMenuStr);
	ListenTo(itsFileMenu);

	JXImage* image = new JXImage(GetDisplay(), GetColormap(), JXPM(filenew));
	assert(image != NULL);
	itsFileMenu->SetItemImage(kNewCmd, image, kTrue);
	toolBar->AppendButton(itsFileMenu, kNewCmd);
	
	image = new JXImage(GetDisplay(), GetColormap(), JXPM(fileopen));
	assert(image != NULL);
	itsFileMenu->SetItemImage(kOpenCmd, image, kTrue);
	toolBar->AppendButton(itsFileMenu, kOpenCmd);
	
	image = new JXImage(GetDisplay(), GetColormap(), JXPM(filefloppy));
	assert(image != NULL);
	itsFileMenu->SetItemImage(kSaveCmd, image, kTrue);
	toolBar->AppendButton(itsFileMenu, kSaveCmd);

	toolBar->NewGroup();
	
	image = new JXImage(GetDisplay(), GetColormap(), JXPM(fileprint));
	assert(image != NULL);	
	itsFileMenu->SetItemImage(kPrintCmd, image, kFalse);
	toolBar->AppendButton(itsFileMenu, kPrintCmd);

    itsScrollbarSet =
        new JXScrollbarSet(window,
                    JXWidget::kHElastic, JXWidget::kVElastic, 
                    0,kJXStdMenuBarHeight + kToolBarHeight, 
                    w, h - kJXStdMenuBarHeight - kToolBarHeight);
    assert( itsScrollbarSet != NULL );

	window->SetMinSize(150,150);

	// layout table and headers

	const JCoordinate kRowHeaderWidth  = 30;
	const JCoordinate kColHeaderHeight = 20;

	JXContainer* encl = itsScrollbarSet->GetScrollEnclosure();
	JRect enclApG     = encl->GetApertureGlobal();

	JXTextButton* okButton = 
		new JXTextButton("Ok", encl,
						JXWidget::kFixedLeft, JXWidget::kFixedTop,
						0, 0, kRowHeaderWidth-2, kColHeaderHeight-2);
	assert(okButton != NULL);
	okButton->SetFontName("6x13");

	itsTable =
		new GXRaggedFloatTable(this, toolBar, okButton, itsData, 6, 
							menuBar, itsScrollbarSet, encl,
						   	JXWidget::kHElastic, JXWidget::kVElastic,
						   	kRowHeaderWidth,kColHeaderHeight,
						   	enclApG.width()  - kRowHeaderWidth,
						   	enclApG.height() - kColHeaderHeight);
	assert( itsTable != NULL );

	enclApG = encl->GetApertureGlobal();	// JXScrollableWidget forces a change in this

	GXRowHeaderWidget* rowHeader =
		GXRowHeaderWidget::Create(itsTable, itsScrollbarSet, encl,
							  JXWidget::kFixedLeft, JXWidget::kVElastic,
							  0,kColHeaderHeight, kRowHeaderWidth,
							  enclApG.height() - kColHeaderHeight);
	assert( rowHeader != NULL );

	GXColHeaderWidget* colHeader =
		GXColHeaderWidget::Create(itsTable, itsScrollbarSet, encl,
							  JXWidget::kHElastic, JXWidget::kFixedTop,
							  kRowHeaderWidth,0, enclApG.width() - kRowHeaderWidth,
							  kColHeaderHeight);
	colHeader->TurnOnColResizing(20);
	assert( colHeader != NULL );
	
	JXDocumentMenu* windowListMenu = 
		new JXDocumentMenu(kWindowListMenuTitleStr, menuBar,
			JXWidget::kFixedLeft, JXWidget::kVElastic, 0,0, 10,10);
	assert( windowListMenu != NULL );
	menuBar->AppendMenu(windowListMenu);
	
	itsExportMenu = new JXTextMenu(itsFileMenu, kExportCmd, menuBar);
	assert( itsExportMenu != NULL );
	itsExportMenu->SetMenuItems(kExportMenuStr);
	itsExportMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsExportMenu);
	
	itsHelpMenu = menuBar->AppendTextMenu(kHelpMenuTitleStr);
	itsHelpMenu->SetMenuItems(kHelpMenuStr);
	itsHelpMenu->SetUpdateAction(JXMenu::kDisableNone);
	ListenTo(itsHelpMenu);

	toolBar->NewGroup();
	
	image = new JXImage(GetDisplay(), GetColormap(), JXPM(manual));
	assert(image != NULL);
	itsHelpMenu->SetItemImage(kTOCCmd, image, kFalse);
	toolBar->AppendButton(itsHelpMenu, kTOCCmd);
	
}

/******************************************************************************
 Receive (protected)

 ******************************************************************************/

void
TestFloatTableDirector::Receive
	(
	JBroadcaster*	sender,
	const Message&	message
	)
{
	if (sender == itsFileMenu && message.Is(JXMenu::kNeedsUpdate))
		{
		UpdateFileMenu();
		}
	else if (sender == itsFileMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleFileMenu(selection->GetIndex());
		}
		
	else if (sender == itsExportMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleExportMenu(selection->GetIndex());
		}

	else if (sender == itsHelpMenu && message.Is(JXMenu::kItemSelected))
		{
		const JXMenu::ItemSelected* selection =
			dynamic_cast(const JXMenu::ItemSelected*, &message);
		assert( selection != NULL );
		HandleHelpMenu(selection->GetIndex());
		}

	else if (sender == itsPrinter &&
			 message.Is(JPrinter::kPrintSetupFinished))
		{
		const JPrinter::PrintSetupFinished* info = 
			dynamic_cast(const JPrinter::PrintSetupFinished*, &message);
		assert(info != NULL);
		if (info->Successful())
			{
			itsTable->PrintRealTable(*itsPrinter);
			}
		}

	else if (sender == itsFileImportDialog && message.Is(JXDialogDirector::kDeactivated))
		{
		const JXDialogDirector::Deactivated* info =
			dynamic_cast(const JXDialogDirector::Deactivated*, &message);
		assert( info != NULL );
		if (info->Successful())
			{
			LoadImportFile();
			}
		itsFileImportDialog = NULL;
		}

	else if (sender == itsDelimiterDialog && message.Is(JXDialogDirector::kDeactivated))
		{
		const JXDialogDirector::Deactivated* info =
			dynamic_cast(const JXDialogDirector::Deactivated*, &message);
		assert( info != NULL );
		if (info->Successful())
			{
			LoadDelimitedFile();
			}
		itsDelimiterDialog = NULL;
		}
		
	else if (sender == itsData && itsListenToData)
		{
		DataModified();
		}
		
	else if (message.Is(J2DPlotWidget::kPlotChanged) ||
			 message.Is(J2DPlotWidget::kTitleChanged) ||
			 message.Is(J2DPlotWidget::kIsEmpty))
		{
		DataModified();
		}

	else
		{
		JXWindowDirector::Receive(sender, message);
		}
}

/******************************************************************************
 UpdateFileMenu (private)

 ******************************************************************************/

void
TestFloatTableDirector::UpdateFileMenu()
{
	itsFileMenu->EnableItem(kOpenCmd);
	itsFileMenu->EnableItem(kNewCmd);
	if (NeedsSave())
		{
		itsFileMenu->EnableItem(kSaveCmd);
		}
	itsFileMenu->EnableItem(kSaveAsCmd);
	itsFileMenu->EnableItem(kPageSetupCmd);
	itsFileMenu->EnableItem(kCloseCmd);
	itsFileMenu->EnableItem(kQuitCmd);

	if (!itsData->IsEmpty())
		{
		itsFileMenu->EnableItem(kPrintCmd);
		}
}

/******************************************************************************
 HandleFileMenu (private)

 ******************************************************************************/

void
TestFloatTableDirector::HandleFileMenu
	(
	const JIndex index
	)
{
	if (index == kNewCmd)
		{
//		itsData->RemoveAllCols();
//		itsData->RemoveAllRows();
		gPlotApp->NewFile();
		}
	else if (index == kOpenCmd)
		{
		JString filename;
		if (JGetChooseSaveFile()->ChooseFile("Select Data File", "", &filename))
			{
			gPlotApp->OpenFile(filename);
			}
		}
	else if (index == kSaveCmd)
		{
		SaveInCurrentFile();
		}
	else if (index == kSaveAsCmd)
		{
		SaveInNewFile();
		}
	else if (index == kPageSetupCmd)
		{
		itsPrinter->BeginUserPageSetup();
		}
	else if (index == kPrintCmd && itsTable->EndEditing())
		{
		itsPrinter->BeginUserPrintSetup();
		}
	else if (index == kCloseCmd)
		{
		Close();
		}
		
	else if (index == kQuitCmd)
		{
		gPlotApp->Quit();
		}
}

/******************************************************************************
 LoadFile (private)

 ******************************************************************************/

void
TestFloatTableDirector::LoadFile
	(
	const JCharacter* filename
	)
{
	ifstream is(filename);
	
	if (is.bad())
		{
		JGetUserNotification()->ReportError("Error opening file.");
		}
	else
		{
		char line[255];
		is.getline(line,255);
		
		JString str(line);
		if (str == kGloveFileSignature)
			{
			LoadNativeFile(is);
			}
			
		else
			{
			is.close();
			itsCurrentFileName = filename;
			ChooseFileFilter();
			}
		}
}

/******************************************************************************
 LoadNativeFile (private)

 ******************************************************************************/

void
TestFloatTableDirector::LoadNativeFile
	(
	ifstream& is
	)
{
	JFloat version;
	is >> version;
	is >> ws;
	
	itsListenToData = kFalse;
	itsTable->ReadData(is, version);
	if (version > 0.5)
		{
		ReadPlotData(is);
		GetWindow()->ReadGeometry(is);
		JXScrollbar* vscroll = itsScrollbarSet->GetVScrollbar();
		vscroll->ReadSetup(is);
		JXScrollbar* hscroll = itsScrollbarSet->GetHScrollbar();
		hscroll->ReadSetup(is);
		}
	itsListenToData = kTrue;
	DataReverted();
}

/******************************************************************************
 ChooseFileFilter (private)

 ******************************************************************************/

void
TestFloatTableDirector::ChooseFileFilter()
{
	assert (itsFileImportDialog == NULL);
	itsFileImportDialog = new GXChooseFileImportDialog(this, itsCurrentFileName);
	assert (itsFileImportDialog != NULL);
	ListenTo(itsFileImportDialog);
	itsFileImportDialog->BeginDialog();
}

/******************************************************************************
 LoadImportFile (private)

 ******************************************************************************/

void
TestFloatTableDirector::LoadImportFile()
{
	JIndex filterIndex = itsFileImportDialog->GetFilterIndex();
	if (filterIndex <= kInternalModuleCount)
		{
		LoadInternalFile(filterIndex);
		}
	else
		{
		JString filter;
		if (!(gPlotApp->GetImportModulePath(filterIndex - kInternalModuleCount, &filter)))
			{
			return;
			}
		
		JString sysCmd = filter + " " + itsCurrentFileName;
		
		int inFD;
		int pid;
		JError err = JExecute(sysCmd, &pid, 
							  kJIgnoreConnection, NULL,
							  kJCreatePipe, &inFD,
							  kJIgnoreConnection, NULL);
							  
		if (!err.OK())
			{
			JGetUserNotification()->ReportError("Error executing filter.");		
			return;
			}

		JInPipeStream ip(inFD, kTrue);
			
		int type;
		ip >> type;
		
		if (type == kGloveDataError)
			{
			JGetUserNotification()->ReportError("This file is not readable by this filter.");
			return;
			}
		
		JSize colCount;
		JFloat value;
	//	itsData->TurnOnBroadcast(kFalse);
		itsListenToData = kFalse;
		if (type == kGloveMatrixDataFormat)
			{
			ip >> colCount;
			for (JSize i = 1; i <= colCount; i++)
				{
				itsData->InsertCol(1);
				}
				
			JBoolean keepGoing = kTrue;	
			JSize rowCount = 0;
			JSize rowChunk = 50;
			JProgressDisplay* pg   = NULL;
			pg = JGetCreatePG()->New();
			pg->VariableLengthProcessBeginning(
				"Loading file...", kTrue, kTrue);
				
			while (!ip.eof() && keepGoing)
				{		
				rowCount++;	
				for (JSize i = 1; i <= colCount; i++)
					{
					ip >> value;
					ip >> ws;
					if (ip.eof() && i != colCount)
						{
						break;
						}
					itsData->AppendElement(i, value);
					}
					
				if (rowChunk == 0)
					{
					keepGoing = pg->IncrementProgress(rowCount);
					rowChunk = 50;
					}
				else
					{
					rowChunk --;
					}
				}
				
			pg->ProcessFinished();
			delete pg;
			}
			
		else if (type == kGloveRaggedDataFormat)
			{
			ip >> colCount;
			for (JSize i = 1; i <= colCount; i++)
				{
				itsData->InsertCol(1);
				}
			JBoolean keepGoing		= kTrue;	
			JSize rowCount 			= 0;
			JSize rowChunk 			= 50;
			JSize rowTotal			= 0;
			JProgressDisplay* pg   	= NULL;
			pg = JGetCreatePG()->New();
			pg->VariableLengthProcessBeginning(
				"Loading file...", kTrue, kTrue);

			JSize colIndex = 1;
			while ((colIndex <= colCount) && keepGoing)
				{
				JSize rowCount;
				ip >> rowCount;

				JSize rowIndex = 1;
				while ((rowIndex <= rowCount) && keepGoing)
					{
					ip >> value;
					ip >> ws;
					itsData->AppendElement(colIndex, value);
					rowIndex++;
					rowTotal++;
					if (rowChunk == 0)
						{
						keepGoing = pg->IncrementProgress(rowTotal);
						rowChunk = 50;
						}
					else
						{
						rowChunk --;
						}
					}
				colIndex++;
				}
			pg->ProcessFinished();
			delete pg;
			}
	//	itsData->TurnOnBroadcast(kTrue);
		itsListenToData = kTrue;
		DataReverted();
	//	itsTable->SyncWithData();
	}
}

/******************************************************************************
 CreateNewPlot

 ******************************************************************************/

void 
TestFloatTableDirector::CreateNewPlot
	(
	const JIndex 		type,
	JArray<JFloat>* 	xCol, 
	JArray<JFloat>* 	x2Col,
	JArray<JFloat>* 	yCol, 
	JArray<JFloat>* 	y2Col,
	JBoolean 			linked, 
	const JCharacter*	label 
	)
{
	JString str = "Plot " + JString(itsPlotNumber);
	itsPlotNumber++;
	
	PlotDir* plotDir = new PlotDir(this, this, GetFileName());
	assert (plotDir != NULL);
	itsPlotWindows->Append(plotDir);
	
	JX2DPlotWidget* plot = plotDir->GetPlot();
	assert( plot != NULL );
	ListenTo(plot);

	AddToPlot(itsPlotWindows->GetElementCount(), type, xCol, x2Col, yCol, y2Col, linked, label);
	
	plot->SetTitle(str);
}
						
						
/******************************************************************************
 AddToPlot

 ******************************************************************************/

void 
TestFloatTableDirector::AddToPlot
	(
	const JIndex 	plotIndex,
	const JIndex 	type,
	JArray<JFloat>* xCol, 
	JArray<JFloat>* x2Col,
	JArray<JFloat>* yCol, 
	JArray<JFloat>* y2Col,
	JBoolean 		linked, 
	const JCharacter* label
	)
{
	assert( itsPlotWindows->IndexValid(plotIndex) );
	
	PlotDir* plotDir = itsPlotWindows->NthElement(plotIndex);
	assert (plotDir != NULL);
	
	JX2DPlotWidget* plot = plotDir->GetPlot();
	assert( plot != NULL );
	ListenTo(plot);
	
	if (type == kDataPlot)
		{
		J2DPlotData* curve;
		if (J2DPlotData::Create(&curve, *xCol, *yCol, linked))
			{
			plot->AddCurve(curve, kTrue, label);
			}
		else
			{
			JGetUserNotification()->ReportError("The data columns must have the same number of elements.");
			return;
			}
		if (x2Col != NULL)
			{
			curve->SetXErrors(*x2Col);
			}
		if (y2Col != NULL)
			{
			curve->SetYErrors(*y2Col);
			}
		}
	else if (type == kVectorPlot)
		{
		J2DVectorData* curve;
		if (J2DVectorData::Create(&curve, *xCol, *yCol, *x2Col, *y2Col, linked))
			{
			plot->AddCurve(curve, kTrue, label);
			}
		else
			{
			JGetUserNotification()->ReportError("The data columns must have the same number of elements.");
			return;
			}
		}
	plotDir->Activate();
	DataModified();
}

/******************************************************************************
 GetPlotNames

 ******************************************************************************/

void 
TestFloatTableDirector::GetPlotNames
	(
	JPtrArray<JString>& names
	)
{
	JString* name;
	
	const JIndex index = itsPlotWindows->GetElementCount();
	
	for (JIndex i = 1; i <= index; i++)
		{
		PlotDir* plotDir = itsPlotWindows->NthElement(i);
		assert (plotDir != NULL);
		
		JX2DPlotWidget* plot = plotDir->GetPlot();
		assert( plot != NULL );
		name = new JString(plot->GetTitle());
		if (name->GetLength() > kMaxPlotTitleSize)	
			{
			name->RemoveSubstring(kMaxPlotTitleSize, name->GetLength());
			name->Append("...");
			}
		names.Append(name);
		}
}

/******************************************************************************
 DirectorClosed

 ******************************************************************************/

void
TestFloatTableDirector::DirectorClosed
	(
	JXDirector* theDirector
	)
{
	JIndex index;

	const PlotDir* dir = 
			dynamic_cast(PlotDir*, theDirector);

	if (itsPlotWindows->Find(dir, &index))
		{
		itsPlotWindows->Remove(dir);
		DataModified();	
		}
}

/******************************************************************************
 WriteFile

 ******************************************************************************/

void
TestFloatTableDirector::WriteTextFile
	(
	ostream& output,
	const JBoolean safetySave
	) 
	const
{
	output << kGloveFileSignature << endl;
	output << kCurrentGloveVersion << endl;
	itsTable->WriteData(output);
	
	JSize plotCount = itsPlotWindows->GetElementCount();
	output << plotCount << " ";
	
	for (JSize i = 1; i <= plotCount; i++)
		{
		PlotDir* plotDir = itsPlotWindows->NthElement(i);
		assert (plotDir != NULL);
		plotDir->WriteSetup(output);
		plotDir->WriteData(output, itsData);
		}
	GetWindow()->WriteGeometry(output);
	JXScrollbar* vscroll = itsScrollbarSet->GetVScrollbar();
	vscroll->WriteSetup(output);
	JXScrollbar* hscroll = itsScrollbarSet->GetHScrollbar();
	hscroll->WriteSetup(output);
}

/******************************************************************************
 ReadPlotData

 ******************************************************************************/

void
TestFloatTableDirector::ReadPlotData
	(
	istream& is
	) 
{
	JSize count;
	is >> count;
	for (JSize i = 1; i <= count; i++)
		{
		PlotDir* plotDir = new PlotDir(this, this, GetFileName());
		assert (plotDir != NULL);
		
		plotDir->ReadSetup(is);
		plotDir->ReadData(is, itsData);
		plotDir->Activate();
		itsPlotWindows->Append(plotDir);
		}
	itsPlotNumber = count + 1;
}

/******************************************************************************
 HandleExportMenu

 ******************************************************************************/

void
TestFloatTableDirector::HandleExportMenu
	(
	const JIndex index
	)
{
	if (index == kReloadModuleCmd)
		{
		UpdateExportMenu();
		}
	else
		{
		JString modName;
		gPlotApp->GetExportModulePath(index - 1, &modName);
		JString tempname = "Untitled.dat";
		JString filename;
		if (JGetChooseSaveFile()->SaveFile(tempname, "Export file as:", "", &filename))
			{
			JString sysCmd = modName + " " + filename;
	
			int inFD;
			int outFD;
			int pid;
			JError err = JExecute(sysCmd, &pid, 
								  kJCreatePipe, &outFD,
								  kJCreatePipe, &inFD,
								  kJIgnoreConnection, NULL);
								  
			if (!err.OK())
				{
				JGetUserNotification()->ReportError("Error executing module.");		
				return;
				}

			JInPipeStream ip(inFD, kTrue);
			assert( ip.good() );
			JOutPipeStream op(outFD, kTrue);
			assert( op.good() );
				
			int type;
			ip >> type;
					
			JSize colCount;
			JFloat value;
			
			if (type == kGloveMatrixDataFormat)
				{
				itsTable->ExportDataMatrix(op);
				}
			else
				{
				itsTable->ExportData(op);
				}
			}
		}
}

/******************************************************************************
 UpdateExportMenu

 ******************************************************************************/

void
TestFloatTableDirector::UpdateExportMenu()
{
	const JSize mCount = itsExportMenu->GetItemCount();
	JSize i;
	for (i = 2; i <= mCount; i++)
		{
		itsExportMenu->DeleteItem(2);
		}

	gPlotApp->ReloadExportModules();
	JPtrArray<JString>* names = gPlotApp->GetExportModules();
	for (i = 1; i <= names->GetElementCount(); i++)
		{
		itsExportMenu->AppendItem(*(names->NthElement(i)));
		}
}

/******************************************************************************
 HandleHelpMenu

 ******************************************************************************/

void
TestFloatTableDirector::HandleHelpMenu
	(
	const JIndex index
	)
{
	if (index == kAboutCmd)
		{
		GetGLApplication()->DisplayAbout();
		}
	else if (index == kTOCCmd)
		{
		JXGetHelpManager()->ShowTOC();
		}
	else if (index == kThisWindowCmd)
		{
		JXGetHelpManager()->ShowSection(kTableHelpName);
		}
	else if (index == kChangesCmd)
		{
		JXGetHelpManager()->ShowSection(kChangeLogName);
		}
	else if (index == kCreditsCmd)
		{
		JXGetHelpManager()->ShowSection(kCreditsName);
		}
}

/******************************************************************************
 DiscardChanges

 ******************************************************************************/

void
TestFloatTableDirector::DiscardChanges()
{

}

/******************************************************************************
 GetInternalModuleName (public)

 ******************************************************************************/

const JCharacter*
TestFloatTableDirector::GetInternalModuleName
	(
	const JIndex index
	)
{
	assert(index <= kInternalModuleCount);
	assert(index > 0);
	return kInternalModuleNames[index - 1];
}

/******************************************************************************
 GetInternalModuleCount (public)

 ******************************************************************************/

JSize
TestFloatTableDirector::GetInternalModuleCount()
{
	return kInternalModuleCount;
}

/******************************************************************************
 LoadInternalFile (private)

 ******************************************************************************/

void
TestFloatTableDirector::LoadInternalFile
	(
	const JIndex index
	)
{
	if (index == kDelimitedText)
		{
		assert(itsDelimiterDialog == NULL);
		itsDelimiterDialog = 
			new GXGetDelimiterDialog(this, itsFileImportDialog->GetFileText());
		assert(itsDelimiterDialog != NULL);
		ListenTo(itsDelimiterDialog);
		itsDelimiterDialog->BeginDialog();
		}
	else if (index == kFixedWidthText)
		{
		cout << "Got the second one..." << endl;
		}
}

/******************************************************************************
 LoadDelimitedFile (private)

 ******************************************************************************/

void
TestFloatTableDirector::LoadDelimitedFile()
{
	assert(itsDelimiterDialog != NULL);
	GXGetDelimiterDialog::DelimiterType type = itsDelimiterDialog->GetDelimiterType();
	JCharacter delim;
	JBoolean useDelim = kFalse;
	if (type == GXGetDelimiterDialog::kChar)
		{
		delim = itsDelimiterDialog->GetCharacter();
		useDelim = kTrue;
		}
	else if (type == GXGetDelimiterDialog::kSpace)			
		{
		delim = ' ';
		}
	else if (type == GXGetDelimiterDialog::kTab)
		{
		delim = '\t';
		}
	ifstream is(itsCurrentFileName);
	if (is.bad())
		{
		JGetUserNotification()->ReportError("Error opening file.");
		return;
		}
	if (itsDelimiterDialog->IsSkippingLines())
		{
		JSize count = itsDelimiterDialog->SkipLineCount();
		for (JSize i = 1; i <= count; i++)
			{
			JReadLine(is);
			}
		}
	JSize row = 0;
	while (is.good())
		{
		JString line = JReadLine(is);
		if (line.IsEmpty())
			{
			continue;
			}
		if (itsDelimiterDialog->HasComments() && line.BeginsWith(itsDelimiterDialog->GetCommentString()))
			{
			continue;
			}
		row++;
		jistrstream(iss, line.GetCString(), line.GetLength());
		JSize col = 1;
		if (type == GXGetDelimiterDialog::kWhiteSpace)
			{
			while (iss.good())
				{
				JFloat value;
				iss >> value;
				if (!iss.fail())
					{
					itsData->SetElement(row, col, value);
					col++;
					}
				else
					{
					if (col == 1)
						{
						row--;
						}
					continue;
					}
				}
			}
		else
			{
			JBoolean found = kTrue;
			while (found)
				{
				JString strVal = JReadUntil(iss, delim, &found);
				JFloat value;
				strVal.ConvertToFloat(&value);
				itsData->SetElement(row, col, value);
				col++;
				}
			}
		}
}